from collections import OrderedDict, ChainMap

# from weakref import WeakKeyDictionary
# from enum import Enum
# from itertools import chain

# from .LoopIR import Alpha_Rename, SubstArgs, LoopIR_Do
# from .configs import reverse_config_lookup, Config
from .new_analysis_core import *

# from .proc_eqv import get_repr_proc

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Simplification Pass to Clean up Analysis Expressions
"""
_fv_cache = WeakKeyDictionary()

def _FV(a):
    fvs = _fv_cache.get(id(a))
    if fvs is not None:
        return fvs
    # otherwise...
    fvs = _A_Free_Vars(a)
    _fv_cache[id(a)] = fvs
    return fvs

def _union_FV(gen):
    fv = set()
    for e in gen:
        fv |= _FV(e)
    return fv

def _A_Free_Vars(a):
    if isinstance(a, A.Var):
        return { a.name }
    elif isinstance(a, (A.Unk,A.Const,A.ConstSym)):
        return set()
    elif isinstance(a, (A.Not,A.USub,A.Definitely,A.Maybe)):
        return _FV(a.arg)
    elif isinstance(a, A.BinOp):
        return _FV(a.lhs) | _FV(a.rhs)
    elif isinstance(a, A.Stride):
        return { (a.name,a.dim) }
    elif isinstance(a, A.LetStrides):
        bound_names = { (a.name,i) for i,_ in enumerate(a.strides) }
        return _union_FV(a.strides) | (_FV(a.body) - bound_names)
    elif isinstance(a, A.Select):
        return _FV(a.cond) | _FV(a.tcase) | _FV(a.fcase)
    elif isinstance(a, (A.ForAll,A.Exists)):
        return _FV(a.arg) - { a.name }
    elif isinstance(a, A.Tuple):
        return _union_FV(a.args)
    elif isinstance(a, A.LetTuple):
        bound_names = set(a.names)
        return _FV(a.rhs) | (_FV(a.body) - bound_names)
    elif isinstance(a, A.Let):
        bound_names = set(a.names)
        return _union_FV(a.rhs) | (_FV(a.body) - bound_names)
    else:
        assert False, f"Bad Case: {type(a)}"
"""


@extclass(A.expr)
def simplify(self):
    return ASimplify(self).result()


class ASimplify:
    def __init__(self, top_level_a):
        self._init_a = top_level_a

        self._const_prop_cache = dict()
        self._const_vals = ChainMap()

        self._fv_cache = dict()

        a = top_level_a
        # constant propagation
        a = self.cprop(a)
        # dead code elimination
        a = self.dcode(a)
        # constant propagation
        a = self.cprop(a)

        self._final_a = a

    def result(self):
        return self._final_a

    def push_cprop(self):
        self._const_vals = self._const_vals.new_child()

    def pop_cprop(self):
        self._const_vals = self._const_vals.parents

    def is_simple(self, a):
        return isinstance(a, (A.Var, A.Unk, A.Const, A.ConstSym, A.Stride))

    def is_simple_tuple(self, a):
        return isinstance(a, (A.Tuple)) and all(self.is_simple(e) for e in a.args)

    # constant propagation
    def cprop(self, a):
        # if a not in self._const_prop_cache:
        #    self._const_prop_cache[a] = self.cprop_helper(a)
        # return self._const_prop_cache[a]
        return self.cprop_helper(a)

    def cprop_helper(self, a):
        if isinstance(a, A.Var):
            val = self._const_vals.get(a.name)
            return val if val is not None else a

        elif isinstance(a, (A.Unk, A.Const, A.ConstSym)):
            return a

        elif isinstance(a, A.Not):
            arg = self.cprop(a.arg)
            if isinstance(arg, A.Const):
                return ABool(not arg.val)
            elif isinstance(arg, A.Unk):
                return arg
            else:
                return type(a)(arg, a.type, a.srcinfo)

        elif isinstance(a, A.USub):
            arg = self.cprop(a.arg)
            if isinstance(arg, A.Const):
                return A.Const(-arg.val, arg.type, arg.srcinfo)
            elif isinstance(arg, A.Unk):
                return arg
            else:
                return type(a)(arg, a.type, a.srcinfo)

        elif isinstance(a, (A.Definitely, A.Maybe)):
            arg = self.cprop(a.arg)
            if isinstance(arg, A.Const):
                return arg
            elif isinstance(arg, A.Unk):
                return ABool(False if isinstance(a, A.Definitely) else True)
            else:
                return type(a)(arg, a.type, a.srcinfo)

        elif isinstance(a, A.BinOp):
            lhs = self.cprop(a.lhs)
            rhs = self.cprop(a.rhs)
            lunk, runk = isinstance(lhs, A.Unk), isinstance(rhs, A.Unk)
            lconst, rconst = isinstance(lhs, A.Const), isinstance(rhs, A.Const)

            # first, operators resulting in a boolean
            if a.type == T.bool:
                if a.op == "and":
                    if lconst:
                        return rhs if lhs.val else lhs
                    elif rconst:
                        return lhs if rhs.val else rhs
                    # fall-through
                elif a.op == "or":
                    if lconst:
                        return lhs if lhs.val else rhs
                    elif rconst:
                        return rhs if rhs.val else lhs
                    # fall-through
                elif a.op == "==>":
                    if lconst:
                        return rhs if lhs.val else ABool(True)
                    elif rconst:
                        return rhs if rhs.val else ANot(lhs)
                    # fall-through
                elif a.op in ("<", ">", "<=", ">=", "=="):
                    if lconst and rconst:
                        return ABool(
                            (lhs.val < rhs.val)
                            if a.op == "<"
                            else (lhs.val > rhs.val)
                            if a.op == ">"
                            else (lhs.val <= rhs.val)
                            if a.op == "<="
                            else (lhs.val >= rhs.val)
                            if a.op == ">="
                            else (lhs.val == rhs.val)
                        )
                    elif lunk or runk:
                        return A.Unk(T.bool, a.srcinfo)
                    # fall-through
                else:
                    assert False, f"bad case: {a.op}"

            else:
                if lconst and rconst:
                    assert a.type.is_indexable()
                    if a.op == "/":
                        pass
                    else:
                        return AInt(
                            (lhs.val + rhs.val)
                            if a.op == "+"
                            else (lhs.val - rhs.val)
                            if a.op == "-"
                            else (lhs.val * rhs.val)
                            if a.op == "*"
                            else (lhs.val % rhs.val)
                        )
                elif lconst:
                    if a.op == "+" and lhs.val == 0:
                        return rhs
                    elif a.op == "-" and lhs.val == 0:
                        return -rhs
                    elif a.op in ("*", "/", "%") and lhs.val == 0:
                        return lhs
                    elif a.op == "*" and lhs.val == 1:
                        return rhs
                    elif a.op == "%" and lhs.val == 1:
                        return lhs
                    # fall through
                elif rconst:
                    if a.op in ("+", "-") and rhs.val == 0:
                        return lhs
                    elif a.op == "*" and rhs.val == 0:
                        return rhs
                    elif a.op in ("*", "/") and rhs.val == 1:
                        return lhs
                    # fall through
                elif lunk or runk:
                    return A.Unk(a.type, a.srcinfo)
                # fall through

            # finally check for a number of other equality tests
            # that we can collapse out
            if a.op == "==" and type(a.lhs) == type(a.rhs):
                if isinstance(a.lhs, (A.Var, A.ConstSym)):
                    if a.lhs.name == a.rhs.name:
                        return ABool(True)
                elif isinstance(a.lhs, A.Stride):
                    if a.lhs.name == a.rhs.name and a.lhs.dim == a.rhs.dim:
                        return ABool(True)

            # catch the fall-throughs
            return A.BinOp(a.op, lhs, rhs, a.type, a.srcinfo)

        elif isinstance(a, A.Stride):
            key = (a.name, a.dim)
            val = self._const_vals.get(key)
            return val if val is not None else a

        elif isinstance(a, A.LetStrides):
            strides = [self.cprop(s) for s in a.strides]
            self.push_cprop()
            if all(self.is_simple(s) for s in strides):
                for i, s in enumerate(strides):
                    self._const_vals[(a.name, i)] = s
                body = self.cprop(a.body)
            else:
                # must mask out to indicate non-propagation here
                for i, _ in enumerate(strides):
                    self._const_vals[(a.name, i)] = None
                body = self.cprop(a.body)
                body = A.LetStrides(a.name, strides, body, a.type, a.srcinfo)
            self.pop_cprop()
            return body

        elif isinstance(a, A.Select):
            cond = self.cprop(a.cond)
            tcase = self.cprop(a.tcase)
            fcase = self.cprop(a.fcase)
            if isinstance(cond, A.Const):
                return tcase if cond.val else fcase
            else:
                return A.Select(cond, tcase, fcase, a.type, a.srcinfo)

        elif isinstance(a, (A.ForAll, A.Exists)):
            self.push_cprop()
            self._const_vals[a.name] = None
            arg = self.cprop(a.arg)
            self.pop_cprop()
            if isinstance(arg, (A.Const, A.Unk)):
                return arg
            else:
                return type(a)(a.name, arg, a.type, a.srcinfo)

        elif isinstance(a, A.Tuple):
            args = [self.cprop(e) for e in a.args]
            return A.Tuple(args, a.type, a.srcinfo)

        elif isinstance(a, A.LetTuple):
            rhs = self.cprop(a.rhs)
            self.push_cprop()
            if self.is_simple_tuple(rhs):
                for nm, e in zip(a.names, rhs.args):
                    self._const_vals[nm] = e
                body = self.cprop(a.body)
            else:
                for nm in a.names:
                    self._const_vals[nm] = None
                body = self.cprop(a.body)
                body = A.LetTuple(a.names, rhs, body, a.type, a.srcinfo)
            self.pop_cprop()
            return body

        elif isinstance(a, A.Let):
            self.push_cprop()
            name_list = []
            rhs_list = []
            for nm, e in zip(a.names, a.rhs):
                rhs = self.cprop(e)
                if self.is_simple(rhs):
                    self._const_vals[nm] = rhs
                else:
                    name_list.append(nm)
                    rhs_list.append(rhs)
                    self._const_vals[nm] = None
            body = self.cprop(a.body)
            self.pop_cprop()
            if len(name_list) == 0:
                return body
            else:
                body = A.Let(name_list, rhs_list, body, a.type, a.srcinfo)
                return body

        else:
            assert False, f"Bad Case: {type(a)}"

    # dead code elimination
    def dcode(self, a):
        return self.dcode_helper(a)

    def dcode_helper(self, a):
        if isinstance(a, (A.Var, A.Unk, A.Const, A.ConstSym, A.Stride)):
            return a
        elif isinstance(a, (A.Not, A.USub, A.Definitely, A.Maybe)):
            arg = self.dcode(a.arg)
            return type(a)(arg, a.type, a.srcinfo)
        elif isinstance(a, A.BinOp):
            lhs = self.dcode(a.lhs)
            rhs = self.dcode(a.rhs)
            return A.BinOp(a.op, lhs, rhs, a.type, a.srcinfo)
        elif isinstance(a, A.LetStrides):
            names = {(a.name, i) for i, _ in enumerate(a.strides)}
            body = self.dcode(a.body)
            bodyFV = self._FV(body)
            if all(nm not in bodyFV for nm in names):
                return body
            else:
                strides = [self.dcode(s) for s in a.strides]
                return A.LetStrides(a.name, strides, body, a.type, a.srcinfo)
        elif isinstance(a, A.Select):
            cond = self.dcode(a.cond)
            tcase = self.dcode(a.tcase)
            fcase = self.dcode(a.fcase)
            return A.Select(cond, tcase, fcase, a.type, a.srcinfo)
        elif isinstance(a, (A.ForAll, A.Exists)):
            arg = self.dcode(a.arg)
            return type(a)(a.name, arg, a.type, a.srcinfo)
        elif isinstance(a, A.Tuple):
            args = [self.dcode(e) for e in a.args]
            return A.Tuple(args, a.type, a.srcinfo)
        elif isinstance(a, A.LetTuple):
            body = self.dcode(a.body)
            bodyFV = self._FV(body)
            if all(nm not in bodyFV for nm in a.names):
                return body
            else:
                rhs = self.dcode(a.rhs)
                return A.LetTuple(a.names, rhs, body, a.type, a.srcinfo)
        elif isinstance(a, A.Let):
            body = self.dcode(a.body)
            FV = self._FV(body)
            name_list = []
            rhs_list = []
            for nm, rhs in reversed(list(zip(a.names, a.rhs))):
                if nm not in FV:
                    pass
                else:
                    rhs = self.dcode(rhs)
                    FV = (FV - {nm}) | self._FV(rhs)
                    name_list.append(nm)
                    rhs_list.append(rhs)

            if len(name_list) == 0:
                return body
            else:
                return A.Let(name_list, rhs_list, body, a.type, a.srcinfo)
        else:
            assert False, f"Bad Case: {type(a)}"

    def _FV(self, a):
        fvs = self._fv_cache.get(id(a))
        if fvs is not None:
            return fvs
        # otherwise...
        fvs = self._A_Free_Vars(a)
        self._fv_cache[id(a)] = fvs
        return fvs

    def _union_FV(self, gen):
        fv = set()
        for e in gen:
            fv |= self._FV(e)
        return fv

    def _A_Free_Vars(self, a):
        if isinstance(a, A.Var):
            return {a.name}
        elif isinstance(a, (A.Unk, A.Const, A.ConstSym)):
            return set()
        elif isinstance(a, (A.Not, A.USub, A.Definitely, A.Maybe)):
            return self._FV(a.arg)
        elif isinstance(a, A.BinOp):
            return self._FV(a.lhs) | self._FV(a.rhs)
        elif isinstance(a, A.Stride):
            return {(a.name, a.dim)}
        elif isinstance(a, A.LetStrides):
            bound_names = {(a.name, i) for i, _ in enumerate(a.strides)}
            return self._union_FV(a.strides) | (self._FV(a.body) - bound_names)
        elif isinstance(a, A.Select):
            return self._FV(a.cond) | self._FV(a.tcase) | self._FV(a.fcase)
        elif isinstance(a, (A.ForAll, A.Exists)):
            return self._FV(a.arg) - {a.name}
        elif isinstance(a, A.Tuple):
            return self._union_FV(a.args)
        elif isinstance(a, A.LetTuple):
            bound_names = set(a.names)
            return self._FV(a.rhs) | (self._FV(a.body) - bound_names)
        elif isinstance(a, A.Let):
            bound_names = set(a.names)
            return self._union_FV(a.rhs) | (self._FV(a.body) - bound_names)
        else:
            assert False, f"Bad Case: {type(a)}"



from collections import defaultdict
from weakref import WeakKeyDictionary

from .LoopIR import LoopIR, LoopIR_Do

_WC_Leaf = {}  # unique object to use as key....


class WeakCache(WeakKeyDictionary):
    def __init__(self):
        self._tuple_dict = WeakKeyDictionary()
        self._dict = WeakKeyDictionary()

    def __contains__(self, key):
        if isinstance(key, (tuple, list)):
            lookup = self._tuple_dict
            for k in key:
                if k not in lookup:
                    return False
                else:
                    lookup = lookup[k]
            return _WC_Leaf in lookup
        else:
            return key in self._dict

    def __getitem__(self, key):
        if isinstance(key, (tuple, list)):
            lookup = self._tuple_dict
            for k in key:
                lookup = lookup[k]
            return lookup[_WC_Leaf]
        else:
            return self._dict[key]

    def __setitem__(self, key, value):
        if isinstance(key, (tuple, list)):
            lookup = self._tuple_dict
            for k in key:
                if k not in lookup:
                    lookup[k] = WeakKeyDictionary()
                lookup = lookup[k]
            lookup[_WC_Leaf] = value
        else:
            self._dict[key] = value


#
# So, what is dependency analysis?
# Or to put it another way, what extensional property(s)
# does dependency analysis guarantee?
#
# Let B be a block of statements,
#     s be a store, and
#     x, y, … be names/symbols.
# Let FV(B) be the set of names that are free in B
#
# Then, first observe that the "meaning" of B is
#
#   Exec[[B]] : (FV(B) -> Value) -> Store -> Store
#
# (note that (FV(B) -> Value) is a valuation/mapping specifying the values
#       of all free variables)
# (further note that Store = (Name -> Maybe Value) is a valuation/mapping
#       of variables that models the heap/store)
#
# Then, (not x DependsOn y in B) for some y in FV(B) implies that
#
#   (Exec[[B]] (env[ y := v1 ]) s)[x] =
#   (Exec[[B]] (env[ y := v2 ]) s)[x]
#
# for all v1, v2
#
# Or in other words, the meaning of B
# w.r.t. its effect on x
# is invariant to the value of y
# when x does not depend on y in B
#


# data-flow dependencies between variable names
class LoopIR_Dependencies(LoopIR_Do):
    def __init__(self, buf_sym, stmts):
        self._buf_sym = buf_sym
        self._lhs = None
        self._depends = defaultdict(set)
        self._alias = dict()

        # If `lhs` is not None, then `lhs` will become dependent
        # on anything read.
        self._lhs = None

        # variables that affect whether or not the
        # currently examined code is even running
        self._context = set()

        # If `control` is True, then anything read will be added
        # to `context`.
        self._control = False

        self.do_stmts(stmts)

    def result(self):
        depends = self._depends[self._buf_sym]
        new = list(depends)
        done = []
        while True:
            if len(new) == 0:
                break
            sym = new.pop()
            done.append(sym)
            d = self._depends[sym]
            depends.update(d)
            new.extend(s for s in d if s not in done)

        return depends

    def do_s(self, s):
        if isinstance(s, (LoopIR.Assign, LoopIR.Reduce)):
            lhs = self._alias.get(s.name, s.name)
            self._lhs = lhs
            self._depends[lhs].add(lhs)
            self._depends[lhs].update(self._context)
            for i in s.idx:
                self.do_e(i)
            self.do_e(s.rhs)
            self._lhs = None
        elif isinstance(s, LoopIR.WriteConfig):
            lhs = (s.config, s.field)
            self._lhs = lhs
            self._depends[lhs].add(lhs)
            self._depends[lhs].update(self._context)
            self.do_e(s.rhs)
            self._lhs = None
        elif isinstance(s, LoopIR.WindowStmt):
            rhs_buf = self._alias.get(s.rhs.name, s.rhs.name)
            self._alias[s.name] = rhs_buf
            self._lhs = rhs_buf
            self._depends[rhs_buf].add(rhs_buf)
            self.do_e(s.rhs)
            self._lhs = None

        elif isinstance(s, LoopIR.If):
            old_context = self._context
            self._context = old_context.copy()

            self._control = True
            self.do_e(s.cond)
            self._control = False

            self.do_stmts(s.body)
            self.do_stmts(s.orelse)

            self._context = old_context

        elif isinstance(s, LoopIR.For):
            old_context = self._context
            self._context = old_context.copy()

            self._control = True
            self._lhs = s.iter
            self._depends[s.iter].add(s.iter)
            self.do_e(s.lo)
            self.do_e(s.hi)
            self._lhs = None
            self._control = False

            self.do_stmts(s.body)

            self._context = old_context

        elif isinstance(s, LoopIR.Call):

            def process_reads():
                # now handle dependencies on buffers that might
                # be read from in the sub-procedure
                # and dependencies on other arguments
                for faa, aa in zip(s.f.args, s.args):
                    if faa.type.is_numeric():
                        maybe_read = self.analyze_eff(s.f.eff, faa.name, read=True)
                    else:
                        maybe_read = True

                    if maybe_read:
                        self.do_e(aa)

                # additionally, we need to handle dependencies
                # on configuration fields
                for ce in s.f.eff.config_reads:
                    name = (ce.config, ce.field)
                    if self._lhs:
                        self._depends[self._lhs].add(name)

            # for every argument that represents a buffer being
            # written to
            for fa, a in zip(s.f.args, s.args):
                maybe_write = fa.type.is_numeric() and self.analyze_eff(
                    s.f.eff, fa.name, write=True
                )
                if maybe_write:
                    name = self._alias.get(a.name, a.name)
                    self._lhs = name
                    self._depends[name].add(name)
                    self._depends[name].update(self._context)
                    process_reads()
                    self._lhs = None

            # secondly, for every configuration field being written to
            # by this sub-procedure, we need to determine dependencies
            for ce in s.f.eff.config_writes:
                name = (ce.config, ce.field)
                self._lhs = name
                self._depends[name].add(name)
                self._depends[name].update(self._context)
                process_reads()
                self._lhs = None

        elif isinstance(s, (LoopIR.Pass, LoopIR.Alloc)):
            pass
        else:
            assert False, "bad case"

    def analyze_eff(self, eff, buf, write=False, read=False):
        if read:
            if any(es.buffer == buf for es in eff.reads):
                return True
        if write:
            if any(es.buffer == buf for es in eff.writes):
                return True
        if read or write:
            if any(es.buffer == buf for es in eff.reduces):
                return True

        return False

    def do_e(self, e):
        if isinstance(e, (LoopIR.Read, LoopIR.WindowExpr)):

            def visit_idx(e):
                if isinstance(e, LoopIR.Read):
                    for i in e.idx:
                        self.do_e(i)
                else:
                    for w in e.idx:
                        if isinstance(w, LoopIR.Interval):
                            self.do_e(w.lo)
                            self.do_e(w.hi)
                        else:
                            self.do_e(w.pt)

            name = self._alias.get(e.name, e.name)
            if self._lhs:
                self._depends[self._lhs].add(name)
            if self._control:
                self._context.add(name)

            visit_idx(e)

        elif isinstance(e, LoopIR.ReadConfig):
            name = (e.config, e.field)
            if self._lhs:
                self._depends[self._lhs].add(name)
            if self._control:
                self._context.add(name)

        else:
            super().do_e(e)

    def do_t(self, t):
        pass

    def do_eff(self, eff):
        pass



from __future__ import annotations

import itertools
import os
import platform

import pytest

from exo import proc
from exo.platforms.neon import *
from exo.stdlib.scheduling import *
from exo.memory import MemGenError

import numpy as np


def test_neon_can_read():
    @proc
    def read_neon(n: size, dst: R[n] @ DRAM, src: R[n] @ Neon):
        for i in seq(0, n):
            dst[i] = src[i]

    with pytest.raises(MemGenError, match="cannot read"):
        read_neon.c_code_str()


@pytest.mark.isa("neon")
def test_neon_memcpy(compiler):
    """
    Compute dst = src
    """

    @proc
    def memcpy_neon(n: size, dst: R[n] @ DRAM, src: R[n] @ DRAM):  # pragma: no cover
        for i in seq(0, (n + 3) / 4):
            if n - 4 * i >= 4:
                tmp: f32[4] @ Neon
                neon_vld_4xf32(tmp, src[4 * i : 4 * i + 4])
                neon_vst_4xf32(dst[4 * i : 4 * i + 4], tmp)
            else:
                for j in seq(0, n - 4 * i):
                    dst[4 * i + j] = src[4 * i + j]

    fn = compiler.compile(
        memcpy_neon, skip_on_fail=True, CMAKE_C_FLAGS="-mcpu=apple-a14"
    )

    for n in (7, 8, 9, 31, 32, 33, 127, 128, 129):
        inp = np.array([float(i) for i in range(n)], dtype=np.float32)
        out = np.array([float(0) for _ in range(n)], dtype=np.float32)
        fn(None, n, out, inp)

        assert np.array_equal(inp, out)


@pytest.mark.isa("neon")
def test_neon_simple_math(compiler):
    """
    Compute x = x * y^2
    """

    @proc
    def simple_math_neon(n: size, x: R[n] @ DRAM, y: R[n] @ DRAM):  # pragma: no cover
        assert n % 4 == 0
        for i in seq(0, n / 4):
            xVec1: f32[4] @ Neon
            xVec2: f32[4] @ Neon
            yVec: f32[4] @ Neon
            neon_vld_4xf32(xVec2, x[4 * i : 4 * i + 4])
            neon_vld_4xf32(yVec, y[4 * i : 4 * i + 4])
            neon_vmul_4xf32(xVec1, xVec2, yVec)
            neon_vmul_4xf32(xVec2, xVec1, yVec)
            neon_vst_4xf32(x[4 * i : 4 * i + 4], xVec2)

    fn = compiler.compile(
        simple_math_neon, skip_on_fail=True, CMAKE_C_FLAGS="-mcpu=apple-a14"
    )

    for n in (4, 8, 12, 16, 24, 32, 64, 128):
        x = np.array([float(i) for i in range(n)], dtype=np.float32)
        y = np.array([float(3 * i) for i in range(n)], dtype=np.float32)
        expected = x * y * y

        fn(None, n, x, y)
        assert np.allclose(x, expected)


@pytest.fixture
def test_neon_vfmla():
    """
    Compute C[i] = A[i] * B[l]
    """

    @proc
    def vfmla(
        n: size, C: R[n] @ DRAM, A: R[n] @ DRAM, B: R[n] @ DRAM
    ):  # pragma: no cover
        assert n == 4
        for l in seq(0, 4):
            for i in seq(0, 4):
                C[i] += A[i] * B[l]

    def simple_vfmla(p=vfmla):
        p = stage_mem(p, "C[_] += _", "C[i]", "C_reg")
        p = expand_dim(p, "C_reg", 4, "i")
        p = lift_alloc(p, "C_reg", n_lifts=2)
        p = autofission(p, p.find("C_reg[_] = _").after(), n_lifts=2)
        p = autofission(p, p.find("C[_] = _").before(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_4xf32)
        p = replace(p, "for i in _: _ #1", neon_vst_4xf32)
        p = set_memory(p, "C_reg", Neon)

        p = bind_expr(p, "A[_]", "A_vec")
        p = expand_dim(p, "A_vec", 4, "i")
        p = lift_alloc(p, "A_vec", n_lifts=2)
        p = autofission(p, p.find("A_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_4xf32)
        p = set_memory(p, "A_vec", Neon)

        p = bind_expr(p, "B[_]", "B_vec")
        p = expand_dim(p, "B_vec", 4, "l")
        p = lift_alloc(p, "B_vec", n_lifts=2)
        p = autofission(p, p.find("B_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for l in _: _ #0", neon_vld_4xf32)
        p = set_memory(p, "B_vec", Neon)

        p = replace(p, "for i in _: _ #0", neon_vfmla_4xf32_4xf32)
        p = unroll_loop(p, "l #0")
        return p

    simple_neon_vfmla = simple_vfmla()

    return simplify(simple_neon_vfmla)


@pytest.mark.isa("neon")
def test_gen_neon_vfmla(golden, test_neon_vfmla):
    assert str(test_neon_vfmla) == golden


@pytest.fixture
def test_neon_vfmla_f16():
    """
    Compute C[i] = A[i] * B[l]
    """

    @proc
    def vfmla_f16(
        n: size, C: R[n] @ DRAM, A: R[n] @ DRAM, B: R[n] @ DRAM
    ):  # pragma: no cover
        assert n == 8
        for l in seq(0, 8):
            for i in seq(0, 8):
                C[i] += A[i] * B[l]

    def simple_vfmla_f16(p=vfmla_f16):
        p = stage_mem(p, "C[_] += _", "C[i]", "C_reg")
        p = expand_dim(p, "C_reg", 8, "i")
        p = lift_alloc(p, "C_reg", n_lifts=2)
        p = autofission(p, p.find("C_reg[_] = _").after(), n_lifts=2)
        p = autofission(p, p.find("C[_] = _").before(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = replace(p, "for i in _: _ #1", neon_vst_8xf16)
        p = set_memory(p, "C_reg", Neon)

        p = bind_expr(p, "A[_]", "A_vec")
        p = expand_dim(p, "A_vec", 8, "i")
        p = lift_alloc(p, "A_vec", n_lifts=2)
        p = autofission(p, p.find("A_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "A_vec", Neon)

        p = bind_expr(p, "B[_]", "B_vec")
        p = expand_dim(p, "B_vec", 8, "l")
        p = lift_alloc(p, "B_vec", n_lifts=2)
        p = autofission(p, p.find("B_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for l in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "B_vec", Neon)

        p = replace(p, "for i in _: _ #0", neon_vfmla_8xf16_8xf16)
        p = unroll_loop(p, "l #0")
        return p

    simple_neon_vfmla_f16 = simple_vfmla_f16()

    return simplify(simple_neon_vfmla_f16)


@pytest.mark.isa("neon")
def test_gen_neon_vfmla_f16(golden, test_neon_vfmla_f16):
    assert str(test_neon_vfmla_f16) == golden


@pytest.fixture
def test_neon_vfmla_f16():
    """
    Compute C[i] = A[i] * B[l]
    """

    @proc
    def vfmla_f16(
        n: size, C: R[n] @ DRAM, A: R[n] @ DRAM, B: R[n] @ DRAM
    ):  # pragma: no cover
        assert n == 8
        for l in seq(0, 8):
            for i in seq(0, 8):
                C[i] += A[i] * B[l]

    def simple_vfmla_f16(p=vfmla_f16):
        p = stage_mem(p, "C[_] += _", "C[i]", "C_reg")
        p = expand_dim(p, "C_reg", 8, "i")
        p = lift_alloc(p, "C_reg", n_lifts=2)
        p = autofission(p, p.find("C_reg[_] = _").after(), n_lifts=2)
        p = autofission(p, p.find("C[_] = _").before(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = replace(p, "for i in _: _ #1", neon_vst_8xf16)
        p = set_memory(p, "C_reg", Neon)

        p = bind_expr(p, "A[_]", "A_vec")
        p = expand_dim(p, "A_vec", 8, "i")
        p = lift_alloc(p, "A_vec", n_lifts=2)
        p = autofission(p, p.find("A_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "A_vec", Neon)

        p = bind_expr(p, "B[_]", "B_vec")
        p = expand_dim(p, "B_vec", 8, "l")
        p = lift_alloc(p, "B_vec", n_lifts=2)
        p = autofission(p, p.find("B_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for l in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "B_vec", Neon)

        p = replace(p, "for i in _: _ #0", neon_vfmla_8xf16_8xf16)
        p = unroll_loop(p, "l #0")
        return p

    simple_neon_vfmla_f16 = simple_vfmla_f16()

    return simplify(simple_neon_vfmla_f16)


@pytest.mark.isa("neon")
def test_gen_neon_vfmla_f16(golden, test_neon_vfmla_f16):
    assert str(test_neon_vfmla_f16) == golden


@pytest.fixture
def test_neon_vfmla_f16():
    """
    Compute C[i] = A[i] * B[l]
    """

    @proc
    def vfmla_f16(
        n: size, C: R[n] @ DRAM, A: R[n] @ DRAM, B: R[n] @ DRAM
    ):  # pragma: no cover
        assert n == 8
        for l in seq(0, 8):
            for i in seq(0, 8):
                C[i] += A[i] * B[l]

    def simple_vfmla_f16(p=vfmla_f16):
        p = stage_mem(p, "C[_] += _", "C[i]", "C_reg")
        p = expand_dim(p, "C_reg", 8, "i")
        p = lift_alloc(p, "C_reg", n_lifts=2)
        p = autofission(p, p.find("C_reg[_] = _").after(), n_lifts=2)
        p = autofission(p, p.find("C[_] = _").before(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = replace(p, "for i in _: _ #1", neon_vst_8xf16)
        p = set_memory(p, "C_reg", Neon)

        p = bind_expr(p, "A[_]", "A_vec")
        p = expand_dim(p, "A_vec", 8, "i")
        p = lift_alloc(p, "A_vec", n_lifts=2)
        p = autofission(p, p.find("A_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for i in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "A_vec", Neon)

        p = bind_expr(p, "B[_]", "B_vec")
        p = expand_dim(p, "B_vec", 8, "l")
        p = lift_alloc(p, "B_vec", n_lifts=2)
        p = autofission(p, p.find("B_vec[_] = _").after(), n_lifts=2)
        p = replace(p, "for l in _: _ #0", neon_vld_8xf16)
        p = set_memory(p, "B_vec", Neon)

        p = replace(p, "for i in _: _ #0", neon_vfmla_8xf16_8xf16)
        p = unroll_loop(p, "l #0")
        return p

    simple_neon_vfmla_f16 = simple_vfmla_f16()

    return simplify(simple_neon_vfmla_f16)


@pytest.mark.isa("neon")
def test_gen_neon_vfmla_f16(golden, test_neon_vfmla_f16):
    assert str(test_neon_vfmla_f16) == golden


@pytest.fixture
def simple_math_neon_sched():
    @proc
    def simple_math_neon_sched(
        n: size, x: R[n] @ DRAM, y: R[n] @ DRAM
    ):  # pragma: no cover
        for i in seq(0, n):
            x[i] = x[i] * y[i] * y[i]

    def sched_neon(p=simple_math_neon_sched):
        p = divide_loop(p, "i", 4, ["io", "ii"], tail="cut_and_guard")

        p = stage_mem(p, "for ii in _:_ #0", "x[4 * io : 4 * io + 4]", "xVec")

        p = bind_expr(p, p.find("y[_]", many=True), "yVec")
        p = autolift_alloc(p, "yVec: _", keep_dims=True)
        p = fission(p, p.find("yVec[_] = _").after())

        p = bind_expr(p, "xVec[_] * yVec[_]", "xy")
        p = autolift_alloc(p, "xy: _", keep_dims=True)
        p = fission(p, p.find("xy[_] = _").after())

        p = set_memory(p, "xVec", Neon)
        p = set_memory(p, "yVec", Neon)
        p = set_memory(p, "xy", Neon)
        p = replace(p, "for i0 in _: _ #1", neon_vst_4xf32)
        p = replace_all(p, neon_vld_4xf32)
        p = replace_all(p, neon_vmul_4xf32)

        p = simplify(p)
        return p

    simple_math_neon_sched = sched_neon()

    return simple_math_neon_sched


def test_gen_neon_simple_math_scheduling(golden, simple_math_neon_sched):
    assert str(simple_math_neon_sched) == golden


@pytest.mark.isa("neon")
def test_neon_simple_math_scheduling(compiler, simple_math_neon_sched):
    """
    Compute x = x * y^2
    """

    fn = compiler.compile(
        simple_math_neon_sched, skip_on_fail=True, CMAKE_C_FLAGS="-mcpu=apple-a14"
    )

    for n in (4, 8, 12, 16, 24, 32, 64, 128):
        x = np.array([float(i) for i in range(n)], dtype=np.float32)
        y = np.array([float(3 * i) for i in range(n)], dtype=np.float32)
        expected = x * y * y

        fn(None, n, x, y)
        assert np.allclose(x, expected)



from __future__ import annotations

from exo import proc, instr, DRAM, config, QAST
from exo.libs.memories import GEMM_SCRATCH, GEMM_ACCUM
from exo.stdlib.scheduling import *


def set_prec_mem(p, bufname, precision, memory):
    p = set_memory(p, bufname, memory)
    p = set_precision(p, bufname, precision)
    return p


old_split = repeat(divide_loop)
old_reorder = repeat(reorder_loops)
old_unroll = repeat(unroll_loop)


def old_fission_after(proc, stmt_pattern, n_lifts=1):
    def find_stmts(p):
        return [c.after() for c in p.find_all(stmt_pattern)]

    return loop_hack(autofission, find_stmts)(proc, n_lifts)


def old_lift_alloc(proc, stmt_pat, n_lifts=1, mode="row", size=None, keep_dims=True):
    def find_stmts(p):
        return p.find_all(stmt_pat)

    return loop_hack(autolift_alloc, find_stmts)(proc, n_lifts, mode, size, keep_dims)


def split_fission_dim(conv):
    conv = old_split(conv, "ocol", 16, ["ocol_o", "ocol_i"], tail="cut_and_guard")
    conv = old_split(conv, "och", 16, ["och_o", "och_i"], perfect=True)
    conv = old_split(conv, "kch", 16, ["kch_o", "kch_i"], perfect=True)
    conv = old_reorder(conv, "ocol_i och_o")
    conv = old_lift_alloc(conv, "res : _", n_lifts=3)
    conv = old_fission_after(conv, "res[_] = _", n_lifts=3)
    conv = old_fission_after(conv, "for krow in _:_", n_lifts=3)
    conv = old_reorder(conv, "och_i krow")
    conv = old_reorder(conv, "och_i kcol")
    conv = old_reorder(conv, "och_i kch_o")
    conv = old_reorder(conv, "ocol_i krow")
    conv = old_reorder(conv, "ocol_i kcol")
    conv = old_reorder(conv, "ocol_i kch_o")
    conv = old_reorder(conv, "och_o krow")
    conv = simplify(conv)
    conv = old_lift_alloc(conv, "i_s : _", n_lifts=6)
    conv = old_lift_alloc(conv, "w_s : _", n_lifts=1)
    conv = old_lift_alloc(conv, "w_s : _", n_lifts=1, mode="col")
    conv = old_reorder(conv, "och_o kcol")
    conv = old_reorder(conv, "och_o kch_o")
    conv = old_lift_alloc(conv, "w_s : _", n_lifts=3)
    conv = old_fission_after(conv, "w_s = _", n_lifts=5)
    conv = old_fission_after(conv, "i_s = _", n_lifts=5)

    return conv


def replace_div_part(conv):
    conv = replace(conv, "for ocol_i in _:_ #0", ld_acc_i32_vector)
    conv = old_reorder(conv, "och_i kch_i")
    conv = old_reorder(conv, "och_o kch_i")
    conv = replace(conv, "for kch_i in _:_ #0", ld_i8_block_id1)
    conv = old_reorder(conv, "kch_o ocol_i")
    conv = replace(conv, "for ocol_i in _:_ #0", ld_i8_block_id2)
    conv = old_reorder(conv, "kch_i och_i")
    conv = replace(conv, "for ocol_i in _:_ #0", matmul_acc_i8)
    conv = replace(conv, "for ocol_i in _:_ #0", st_acc_i8)

    return conv


def replace_mod_part(conv):
    conv = replace(conv, "for ocol_i in _:_ #0", ld_acc_i32_vector)
    conv = old_reorder(conv, "och_i kch_i")
    conv = replace(conv, "for kch_i in _:_ #0", ld_i8_block_id1)
    conv = replace(conv, "for ocol_i in _:_ #0", ld_i8_block_id2)
    conv = old_reorder(conv, "kch_i och_i")
    conv = replace(conv, "for ocol_i in _:_ #0", matmul_acc_i8)
    conv = replace(conv, "for ocol_i in _:_ #0", st_acc_i8)

    return conv


def matmul_tile(gemmini):
    gemmini = divide_loop(gemmini, "j", 4, ["jo", "ji"], perfect=True)
    gemmini = divide_loop(gemmini, "i", 8, ["io", "i"], perfect=True)
    gemmini = divide_loop(gemmini, "io", 2, ["ioo", "io"], perfect=True)
    gemmini = old_reorder(gemmini, "i jo")
    gemmini = old_reorder(gemmini, "io jo")
    return gemmini


def inline_lift_config(gemmini):
    # part of scheduling count, 25
    gemmini = call_eqv(gemmini, "zero_acc_i32(_, _, _)", zero_acc_i32_v2)
    gemmini = inline(gemmini, "zero_acc_i32_v2(_, _, _)")
    gemmini = inline_window(gemmini, "dst = res[_]")
    gemmini = lift_config(gemmini, "config_zero()")

    gemmini = call_eqv(gemmini, "ld_i8_block_id1(_)", ld_i8_block_id1_v2)
    gemmini = inline(gemmini, "ld_i8_block_id1_v2(_, _, _, _, _)")
    gemmini = inline_window(gemmini, "src = A[_]")
    gemmini = inline_window(gemmini, "dst = a[_]")
    gemmini = lift_config(gemmini, "config_ld_i8_id1()")

    gemmini = call_eqv(gemmini, "ld_i8_block_id2(_)", ld_i8_block_id2_v2)
    gemmini = inline(gemmini, "ld_i8_block_id2_v2(_, _, _, _, _)")
    gemmini = inline_window(gemmini, "src = B[_]")
    gemmini = inline_window(gemmini, "dst = b[_]")
    gemmini = lift_config(gemmini, "config_ld_i8_id2()")

    gemmini = call_eqv(gemmini, "matmul_acc_i8(_, _, _, _, _)", matmul_acc_i8_v2)
    gemmini = inline(gemmini, "matmul_acc_i8_v2(_, _, _, _, _)")
    gemmini = inline_window(gemmini, "A = a[_]")
    gemmini = inline_window(gemmini, "B = b[_]")
    gemmini = inline_window(gemmini, "C = res[_]")
    gemmini = lift_config(gemmini, "config_matmul()")

    gemmini = call_eqv(gemmini, "st_acc_i8(_, _, _, _, _, _)", st_acc_i8_v2)
    gemmini = inline(gemmini, "st_acc_i8_v2(_, _, _, _, _, _)")
    gemmini = inline_window(gemmini, "src = res[_]")
    gemmini = inline_window(gemmini, "dst = C[_]")
    gemmini = lift_config(gemmini, "config_st_acc_i8(_)")
    return gemmini


def replace_gemmini_calls(gemmini):
    gemmini = replace(gemmini, "for i_in in _:_ #0", zero_acc_i32)
    gemmini = replace(gemmini, "for i_in in _:_ #0", ld_i8_block_id1)
    gemmini = replace(gemmini, "for ki in _:_ #0", ld_i8_block_id2)
    gemmini = replace(gemmini, "for i_in in _:_ #0", matmul_acc_i8)
    gemmini = replace(gemmini, "for i_in in _:_ #0", st_acc_i8)
    return gemmini


def fission_inner_blocks(gemmini):
    gemmini = divide_loop(gemmini, "k", 64, ["ko", "k"], perfect=True)
    gemmini = divide_loop(gemmini, "k", 16, ["k", "ki"], perfect=True)
    gemmini = old_lift_alloc(gemmini, "a : i8", n_lifts=3)
    gemmini = old_lift_alloc(gemmini, "a : _ #0", n_lifts=1, mode="col")
    gemmini = old_lift_alloc(gemmini, "a : _", n_lifts=2)
    gemmini = old_reorder(gemmini, "ki j_in_o")
    gemmini = old_reorder(gemmini, "ki j_in_i")
    gemmini = old_lift_alloc(gemmini, "b : i8", n_lifts=2)
    gemmini = old_lift_alloc(gemmini, "b : i8", n_lifts=1, mode="col")
    gemmini = old_lift_alloc(gemmini, "b : _", n_lifts=3)
    gemmini = old_fission_after(gemmini, "a[_] = _", n_lifts=5)
    gemmini = old_fission_after(gemmini, "b[_] = _", n_lifts=5)
    gemmini = old_reorder(gemmini, "j_in_i i_in")
    gemmini = old_reorder(gemmini, "ki i_in")
    gemmini = old_reorder(gemmini, "k i_in")
    gemmini = old_reorder(gemmini, "j_in_i ki")
    gemmini = old_reorder(gemmini, "j_in_o ki")
    gemmini = old_reorder(gemmini, "j_in_i i_in")
    return gemmini


def inline_div_part(conv):
    conv = inline_vector(conv)
    conv = lift_config(conv, "config_ld_acc_i32_vector(_)")
    conv = inline_ld_id1(conv)
    conv = lift_config(conv, "config_ld_i8_id1(_)")
    conv = inline_matmul(conv)
    conv = lift_config(conv, "config_matmul(_)")
    conv = inline_st(conv)
    conv = lift_config(conv, "config_st_acc_i8(_)")

    return conv


def inline_mod_part(conv):
    conv = inline_vector(conv)
    conv = inline_ld_id1(conv)
    conv = inline_matmul(conv)
    conv = inline_st(conv)
    conv = delete_config(conv, "config_ld_acc_i32_vector(_) #1")
    conv = delete_config(conv, "config_ld_i8_id1(_) #1")
    conv = delete_config(conv, "config_matmul(_) #1")
    conv = delete_config(conv, "config_st_acc_i8(_) #1")
    conv = simplify(conv)

    return conv


def set_gemm_memories(conv):
    conv = set_memory(conv, "res", GEMM_ACCUM)
    conv = set_memory(conv, "i_s", GEMM_SCRATCH)
    conv = set_memory(conv, "w_s", GEMM_SCRATCH)
    try:
        conv = set_memory(conv, "res #1", GEMM_ACCUM)
        conv = set_memory(conv, "i_s #1", GEMM_SCRATCH)
        conv = set_memory(conv, "w_s #1", GEMM_SCRATCH)
    except:
        pass

    return conv


def fission_outer_blocks(gemmini):
    gemmini = old_fission_after(gemmini, "res[_] = 0.0 #0", n_lifts=3)
    gemmini = old_fission_after(gemmini, "for k in _:_ #0", n_lifts=3)
    gemmini = old_reorder(gemmini, "j_in_i j_in_o")
    gemmini = old_reorder(gemmini, "i_in k")
    gemmini = old_reorder(gemmini, "j_in_i k")
    gemmini = old_reorder(gemmini, "j_in_o k")
    return gemmini


def tile_outer_loops(gemmini):
    gemmini = divide_loop(gemmini, "i", 16, ["i", "i_in"], perfect=True)
    gemmini = old_reorder(gemmini, "i_in j")
    gemmini = divide_loop(gemmini, "j", 64, ["j", "j_in"], perfect=True)
    gemmini = divide_loop(gemmini, "j_in", 16, ["j_in_o", "j_in_i"], perfect=True)
    gemmini = old_reorder(gemmini, "j_in_o j_in_i")

    return gemmini


class QAST_Do:
    def __init__(self, proc):
        self.proc = proc

        # [ self.do_fnarg(a) for a in self.proc.args ]
        [self.do_e(p) for p in self.proc.assertions]
        self.do_stmts(self.proc.body)

    def do_stmts(self, stmts):
        [self.do_s(b) for b in stmts]

    def do_s(self, s):
        if type(s) is QAST.Assign or type(s) is QAST.Reduce:
            [self.do_e(e) for e in s.idx]
            self.do_e(s.rhs)
        elif type(s) is QAST.WriteConfig:
            self.do_e(s.rhs)
        elif type(s) is QAST.For:
            self.do_e(s.lo)
            self.do_e(s.hi)
            self.do_stmts(s.body)
        elif type(s) is QAST.If:
            self.do_e(s.cond)
            self.do_stmts(s.body)
            if len(s.orelse) > 0:
                self.do_stmts(s.orelse)
        elif type(s) is QAST.Pass:
            pass
        elif type(s) is QAST.Alloc:
            pass
        elif type(s) is QAST.Call:
            [self.do_e(e) for e in s.args]
        elif type(s) is QAST.WindowStmt:
            self.do_e(s.rhs)
        else:
            assert False, "bad case"

    def do_w_access(self, w):
        if type(w) is QAST.Interval:
            self.do_e(w.lo)
            self.do_e(w.hi)
        elif type(w) is QAST.Point:
            self.do_e(w.pt)

    def do_e(self, e):
        if type(e) is QAST.Read:
            [self.do_e(ei) for ei in e.idx]
        elif type(e) is QAST.Const:
            pass
        elif type(e) is QAST.USub:
            self.do_e(e.arg)
        elif type(e) is QAST.BinOp:
            self.do_e(e.lhs)
            self.do_e(e.rhs)
        elif type(e) is QAST.BuiltIn:
            [self.do_e(ei) for ei in e.args]
        elif type(e) is QAST.WindowExpr:
            [self.do_w_access(w) for w in e.idx]
        elif type(e) is QAST.StrideExpr:
            pass
        elif type(e) is QAST.ReadConfig:
            pass
        else:
            assert False, "bad case"


class CanFissionLoop(QAST_Do):
    def __init__(self, proc, stmt):
        self.stmt = stmt
        self.result = False
        super().__init__(proc)

    def result(self):
        return self.result

    def do_s(self, s):
        if type(s) is QAST.For:
            assert len(s.body) > 0
            if s.body[0] == self.stmt:
                self.result = True

        super().do_s(s)


class CanFissionIf(QAST_Do):
    def __init__(self, proc, stmt):
        self.stmt = stmt
        self.result = None
        super().__init__(proc)

    def result(self):
        return self.result

    def do_s(self, s):
        if type(s) is QAST.If:
            assert len(s.body) > 0
            if s.body[0] == self.stmt:
                self.result = str(s)
            elif len(s.orelse) > 0 and s.orelse[0] == self.stmt:
                self.result = str(s)

        super().do_s(s)


class CanReorder(QAST_Do):
    def __init__(self, proc, stmt):
        self.stmt = stmt
        self.result = None
        super().__init__(proc)

    def result(self):
        return self.result

    def do_stmts(self, stmts):
        prev = None
        for b in stmts:
            if b == self.stmt and prev is not None:
                self.result = str(prev)
            else:
                self.do_s(b)
                prev = b


def lift_config(conv, string, nth=0):
    stmt = conv.get_ast(string)
    stmt = stmt[nth]  # Get the match

    while True:
        proc = conv.get_ast()
        fission_loop = CanFissionLoop(proc, stmt).result
        reorder = CanReorder(proc, stmt).result
        if fission_loop:
            conv = old_fission_after(conv, string)
        elif reorder is not None:
            conv = reorder_stmts(conv, conv.find(string).expand(1, 0))
            # conv = conv.reorder_before(string)
        else:
            break

    return conv


def inline_vector(conv):
    conv = call_eqv(conv, "ld_acc_i32_vector(_)", ld_acc_i32_vector_v2)
    conv = inline(conv, "ld_acc_i32_vector_v2(_)")
    conv = inline_window(conv, "src = bias[_]")
    conv = inline_window(conv, "dst = res[_]")
    return conv


def inline_ld_id1(conv):
    conv = call_eqv(conv, "ld_i8_block_id1(_)", ld_i8_block_id1_s2_v2)
    conv = inline(conv, "ld_i8_block_id1_s2_v2(_)")
    conv = inline_window(conv, "src = weights[_]")
    conv = inline_window(conv, "dst = w_s[_]")
    return conv


def inline_matmul(conv):
    conv = call_eqv(conv, "matmul_acc_i8(_)", matmul_acc_i8_v2)
    conv = inline(conv, "matmul_acc_i8_v2(_)")
    conv = inline_window(conv, "A = i_s[_]")
    conv = inline_window(conv, "B = w_s[_]")
    conv = inline_window(conv, "C = res[_]")
    return conv


def inline_st(conv):
    conv = call_eqv(conv, "st_acc_i8(_)", st_acc_i8_s2_v2)
    conv = inline(conv, "st_acc_i8_s2_v2(_)")
    conv = inline_window(conv, "src = res[_]")
    conv = inline_window(conv, "dst = output[_]")
    return conv


# --------------------------------------------------------------------------- #
#   Instructions
# --------------------------------------------------------------------------- #


@instr("{dst}[0] = ACC_SCALE({src}[0], {scale}[0]);")
def acc_scale(src: i32, dst: f32, scale: f32):
    dst = scale * src


@config
class ConfigLoad:
    src_stride: stride


@config
class ConfigLoad_id1:
    src_stride: stride


@config
class ConfigLoad_id2:
    src_stride: stride


@instr("gemmini_extended3_config_ld({src_stride}, 1.0f, 0, 0);\n")
def config_ld_i8(src_stride: stride):
    ConfigLoad.src_stride = src_stride


@instr("gemmini_extended3_config_ld({src_stride}, 1.0f, 0, 1);\n")
def config_ld_i8_id1(src_stride: stride):
    ConfigLoad_id1.src_stride = src_stride


@instr("gemmini_extended3_config_ld({src_stride}, 1.0f, 0, 2);\n")
def config_ld_i8_id2(src_stride: stride):
    ConfigLoad_id2.src_stride = src_stride


@proc
def ld_i8_prototype(
    n: size,
    m: size,
    src: [i8][n, m] @ DRAM,
    dst: [i8][n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 16
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i, j]


def make_do_ld_i8(name, instr_str, assert_str=None, add_pass=False, p=ld_i8_prototype):
    p = make_instr(p, instr_str)
    if assert_str:
        p = p.add_assertion(
            assert_str, configs=[ConfigLoad, ConfigLoad_id1, ConfigLoad_id2]
        )
    if add_pass:
        p = insert_pass(p, p.find("for i in _: _").before())
    p = rename(p, name)
    return p


do_ld_i8 = make_do_ld_i8(
    "do_ld_i8",
    ("gemmini_extended_mvin( " "&{src_data}, ((uint64_t) &{dst_data}), {m}, {n} );"),
    "stride(src,0) == ConfigLoad.src_stride",
)
do_ld_i8_id1 = make_do_ld_i8(
    "do_ld_i8_id1",
    ("gemmini_extended_mvin2( " "&{src_data}, ((uint64_t) &{dst_data}), {m}, {n} );"),
    "stride(src,0) == ConfigLoad_id1.src_stride",
)
do_ld_i8_id2 = make_do_ld_i8(
    "do_ld_i8_id2",
    ("gemmini_extended_mvin3( " "&{src_data}, ((uint64_t) &{dst_data}), {m}, {n} );"),
    "stride(src,0) == ConfigLoad_id2.src_stride",
)

_gemm_ld_i8 = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, 1.0f, 0, 0);\n"
    "gemmini_extended_mvin( "
    "&{src_data}, ((uint64_t) &{dst_data}), {m}, {n} );"
)
ld_i8 = make_do_ld_i8(
    "ld_i8",
    _gemm_ld_i8,
    add_pass=True,
)

_gemm_ld_i8_block = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, "
    + "1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( &{src_data}, "
    + "((uint64_t) &{dst_data}), 16*{m}, {n} );"
)


@instr(_gemm_ld_i8_block)
def ld_i8_block(
    n: size,
    m: size,
    src: [i8][n, 16 * m] @ DRAM,
    dst: [i8][m, n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 4
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            for k in seq(0, 16):
                dst[j, i, k] = src[i, 16 * j + k]


def make_do_ld_i8_block(name, instr_str, p=ld_i8_block):
    p = rename(p, name)
    p = make_instr(p, instr_str)
    return p


_do_gemm_ld_i8_block_id1 = (
    "gemmini_extended_mvin2( &{src_data}, " + "((uint64_t) &{dst_data}), 16*{m}, {n} );"
)
_do_gemm_ld_i8_block_id2 = (
    "gemmini_extended_mvin3( &{src_data}, " + "((uint64_t) &{dst_data}), 16*{m}, {n} );"
)
do_ld_i8_block_id1 = make_do_ld_i8_block("do_ld_i8_block_id1", _do_gemm_ld_i8_block_id1)
do_ld_i8_block_id2 = make_do_ld_i8_block("do_ld_i8_block_id2", _do_gemm_ld_i8_block_id2)


_gemm_ld_i8_block_id1 = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, "
    + "1.0f, 0, 1);\n"
    + "gemmini_extended_mvin2( &{src_data}, "
    + "((uint64_t) &{dst_data}), 16*{m}, {n} );"
)
_gemm_ld_i8_block_id2 = (
    "gemmini_extended4_config_ld({src}.strides[0]*1, "
    + "1.0f, 0, {n}, 2);\n"
    + "gemmini_extended_mvin3( &{src_data}, "
    + "((uint64_t) &{dst_data}), 16*{m}, {n} );"
)


def make_ld_i8_block(name, ld_id, p=ld_i8_block, stride_val=None):
    if ld_id == 1:
        ConfigLoad = ConfigLoad_id1
        do_ld_i8_block = do_ld_i8_block_id1
        config_ld_i8 = config_ld_i8_id1
        write_stride = "ConfigLoad_id1.src_stride = _"
        _gemm_ld_i8_block = _gemm_ld_i8_block_id1
    else:
        assert ld_id == 2
        ConfigLoad = ConfigLoad_id2
        do_ld_i8_block = do_ld_i8_block_id2
        config_ld_i8 = config_ld_i8_id2
        write_stride = "ConfigLoad_id2.src_stride = _"
        _gemm_ld_i8_block = _gemm_ld_i8_block_id2

    if stride_val:
        p = write_config(p, p.body().before(), ConfigLoad, "src_stride", stride_val)
        p = replace(p, "for i in _:_", do_ld_i8_block)
        p = replace(p, write_stride, config_ld_i8)

    p = rename(p, name)
    p = make_instr(p, _gemm_ld_i8_block)
    return p


ld_i8_block_id1 = make_ld_i8_block("ld_i8_block_id1", 1)
ld_i8_block_id1_v2 = make_ld_i8_block(
    "ld_i8_block_id1_v2", 1, stride_val="stride(src,0)"
)
ld_i8_block_id1_s2_v2 = make_ld_i8_block(
    "ld_i8_block_id1_s2_v2", 1, stride_val="stride(src,0)"
)

ld_i8_block_id2 = make_ld_i8_block("ld_i8_block_id2", 2)
ld_i8_block_id2_v2 = make_ld_i8_block(
    "ld_i8_block_id2_v2", 2, stride_val="stride(src,0)"
)
ld_i8_block_id2_s2_v2 = make_ld_i8_block(
    "ld_i8_block_id2_s2_v2", 2, stride_val="stride(src,0)"
)


_gemm_zero_block_id2 = (
    "gemmini_extended4_config_ld(0, 1.0f, 0, {n}, 2);\n"
    + "gemmini_extended_mvin3( 0, ((uint64_t) &{dst_data}), 16*{m}, {n} );"
)


@instr(_gemm_zero_block_id2)
def zero_block_id2(
    n: size,
    m: size,
    dst: [i8][m, n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 4
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            for k in seq(0, 16):
                dst[j, i, k] = 0.0


_gemm_ld_i8_id1 = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, 1.0f, 0, 1);\n"
    + "gemmini_extended_mvin2( &{src_data}, "
    + "((uint64_t) &{dst_data}), {m}, {n} );"
)
_gemm_ld_i8_id2 = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, 1.0f, 0, 2);\n"
    + "gemmini_extended_mvin3( &{src_data}, "
    + "((uint64_t) &{dst_data}), {m}, {n} );"
)


def make_ld_i8(name, ld_id=0, p=ld_i8, sval=None):
    if ld_id == 0:
        cfg = ConfigLoad
        do_load = do_ld_i8
        cfg_load = config_ld_i8
        wr_stride = "ConfigLoad.src_stride = _"
        instr_str = _gemm_ld_i8
    elif ld_id == 1:
        cfg = ConfigLoad_id1
        do_load = do_ld_i8_id1
        cfg_load = config_ld_i8_id1
        wr_stride = "ConfigLoad_id1.src_stride = _"
        instr_str = _gemm_ld_i8_id1
    else:
        assert ld_id == 2
        cfg = ConfigLoad_id2
        do_load = do_ld_i8_id2
        cfg_load = config_ld_i8_id2
        wr_stride = "ConfigLoad_id2.src_stride = _"
        instr_str = _gemm_ld_i8_id2

    if sval:
        p = write_config(p, p.body().before(), cfg, "src_stride", sval)
        # p = p.configwrite_after('pass', cfg, 'src_stride', sval)
        p = replace(p, "for i in _:_", do_load)
        p = replace(p, wr_stride, cfg_load)

    p = delete_pass(p)
    p = rename(p, name)
    p = make_instr(p, instr_str)
    return p


ld_i8_id1 = make_ld_i8("ld_i8_id1", 1)
ld_i8_id1_v2 = make_ld_i8("ld_i8_id1_v2", 1, sval="stride(src, 0)")
ld_i8_id1_s2_v2 = make_ld_i8("ld_i8_id1_s2_v2", 1, sval="stride(src, 0)")

ld_i8_id2 = make_ld_i8("ld_i8_id2", 2)
ld_i8_id2_v2 = make_ld_i8("ld_i8_id2_v2", 2, sval="stride(src, 0)")
ld_i8_id2_s2_v2 = make_ld_i8("ld_i8_id2_s2_v2", 2, sval="stride(src, 0)")

ld_i8_v2 = make_ld_i8("ld_i8_v2", 0, sval="stride(src,0)")

# finally remove the pass from ld_i8
ld_i8 = make_ld_i8("ld_i8", 0)


_gemm_ld_i8_stride_2 = (
    "gemmini_extended3_config_ld({src}.strides[0]*2, "
    + "1.0f, 0, 1);\n"
    + "gemmini_extended_mvin2( &{src_data}, "
    + "((uint64_t) &{dst_data}), {m}, {n} );"
)


@instr(_gemm_ld_i8_stride_2)
def ld_i8_s2(
    n: size,
    m: size,
    src: [i8][n * 2 - 1, m] @ DRAM,
    dst: [i8][n, 16] @ GEMM_SCRATCH,
):
    assert 0 < n and n <= 16
    assert 0 < m and m <= 16
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i * 2, j]


_gemm_config_ld_i8_id1 = (
    "gemmini_extended3_config_ld({src_stride}*2, " + "1.0f, 0, 1);\n"
)


@instr(_gemm_config_ld_i8_id1)
def config_ld_i8_s2_id1(src_stride: stride):
    ConfigLoad_id1.src_stride = src_stride


_do_gemm_ld_i8_stride_2 = (
    "gemmini_extended_mvin2( &{src_data}, " + "((uint64_t) &{dst_data}), {m}, {n} );"
)


@instr(_do_gemm_ld_i8_stride_2)
def do_ld_i8_s2_id1(
    n: size,
    m: size,
    src: [i8][n * 2 - 1, m] @ DRAM,
    dst: [i8][n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 16
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i * 2, j]


_gemm_ld_i8_vec = (
    "gemmini_extended3_config_ld(1, 1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( &{src_data}, "
    + "((uint64_t) &{dst_data}), 16, 1);"
)


@instr(_gemm_ld_i8_vec)
def ld_i8_vector(
    src: [i8][16] @ DRAM,
    dst: [i8][16] @ GEMM_SCRATCH,
):
    assert stride(dst, 0) == 16

    for i in seq(0, 16):
        dst[i] = src[i]


_do_gemm_ld_i8_vec = (
    "gemmini_extended_mvin( &{src_data}, ((uint64_t) &{dst_data}), 16, 1);"
)


@instr(_do_gemm_ld_i8_vec)
def do_ld_i8_vector(
    src: [i8][16] @ DRAM,
    dst: [i8][16] @ GEMM_SCRATCH,
):
    assert stride(dst, 0) == 16

    for i in seq(0, 16):
        dst[i] = src[i]


# in order to load i8 values into the i32 accumulator memory,
# we must specify `shrunk=1` (3rd param of ..._config_ld)
_gemm_ld_acc_i8 = (
    "gemmini_extended3_config_ld({src}.strides[0]*1, "
    + "1.0f, 1, 0);\n"
    + "gemmini_extended_mvin( &{src_data}, "
    + "((uint32_t) &{dst_data}), {m}, {n} );"
)
ld_acc_i8 = rename(ld_i8, "ld_acc_i8")
ld_acc_i8 = set_prec_mem(ld_acc_i8, "dst", "i32", GEMM_ACCUM)
ld_acc_i8 = make_instr(ld_acc_i8, _gemm_ld_acc_i8)


def new_config_ld_acc():
    @config
    class ConfigLoadAcc:
        stride_set: bool

    return ConfigLoadAcc


ConfigLoadAcc = new_config_ld_acc()

_gemm_ld_acc_i32 = (
    "gemmini_extended3_config_ld({src}.strides[0]*4, "
    + "1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( ((uint64_t) &{src_data}), "
    + "((uint32_t) &{dst_data}), {m}, {n} );"
)


@instr(_gemm_ld_acc_i32)
def ld_acc_i32(
    n: size,
    m: size,
    src: [i32][n, m] @ DRAM,
    dst: [i32][n, 16] @ GEMM_ACCUM,
):
    assert n <= 16
    assert m <= 16
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i, j]


_gemm_do_ld_acc_i32 = (
    "gemmini_extended_mvin( ((uint64_t) &{src_data}), "
    + "((uint32_t) &{dst_data}), {m}, {n} );"
)


@instr(_gemm_do_ld_acc_i32)
def do_ld_acc_i32(
    n: size,
    m: size,
    src: [i32][n, m] @ DRAM,
    dst: [i32][n, 16] @ GEMM_ACCUM,
):
    assert n <= 16
    assert m <= 16
    assert stride(src, 1) == 1
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i, j]


_gemm_config_ld_acc_i32_vector = "gemmini_extended3_config_ld(0, 1.0f, 0, 0);\n"


@instr(_gemm_config_ld_acc_i32_vector)
def config_ld_acc_i32_vector(stride_set: bool):
    ConfigLoadAcc.stride_set = stride_set


_gemm_ld_acc_i32_vec = (
    "gemmini_extended3_config_ld(0, 1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( ((uint64_t) &{src_data}), ((uint32_t) &{dst_data}), 16, {n} );"
)


@instr(_gemm_ld_acc_i32_vec)
def ld_acc_i32_vector(
    n: size,
    src: [i32][1, 16] @ DRAM,
    dst: [i32][n, 16] @ GEMM_ACCUM,
):
    assert n <= 16
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, n):
        for j in seq(0, 16):
            dst[i, j] = src[0, j]


_do_gemm_ld_acc_i32_vec = "gemmini_extended_mvin( ((uint64_t) &{src_data}), ((uint32_t) &{dst_data}), 16, {n} );"


@instr(_do_gemm_ld_acc_i32_vec)
def do_ld_acc_i32_vector(
    n: size,
    src: [i32][1, 16] @ DRAM,
    dst: [i32][n, 16] @ GEMM_ACCUM,
):
    assert n <= 16
    assert stride(dst, 0) == 1
    assert stride(src, 0) == 1

    for i in seq(0, n):
        for j in seq(0, 16):
            dst[i, j] = src[0, j]


def make_ld_acc_i32_vector(name, p=ld_acc_i32_vector):
    p = rename(p, name)
    p = write_config(p, p.body().before(), ConfigLoadAcc, "stride_set", "True")
    # p = p.configwrite_after('pass', ConfigLoadAcc, 'stride_set', 'True')
    p = replace(p, "for i in _:_", do_ld_acc_i32_vector)
    p = replace(p, "ConfigLoadAcc.stride_set = _", config_ld_acc_i32_vector)
    p = make_instr(p, _gemm_ld_acc_i32_vec)
    return p


ld_acc_i32_vector_v2 = make_ld_acc_i32_vector("ld_acc_i32_vector_v2")


_gemm_st_i8 = (
    "gemmini_extended_config_st({dst}.strides[0]*1, 0, 1.0f);\n"
    + "gemmini_extended_mvout( "
    + "((uint64_t) &{dst_data}), (uint32_t) &{src_data}, {m}, {n} );"
)


@instr(_gemm_st_i8)
def st_i8(n: size, m: size, src: [i8][n, 16] @ GEMM_SCRATCH, dst: [i8][n, m] @ DRAM):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 1) == 1
    assert stride(src, 0) == 16
    assert stride(src, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i, j]


@proc
def clamp(src: f32, dst: i8):
    l: f32
    h: f32
    l = -128.0
    h = 127.0
    dst = select(h, src, h, src)
    dst = select(src, l, l, dst)


def new_config_st():
    @config
    class ConfigStore:
        scale: f32
        dst_stride: stride
        act: bool

    return ConfigStore


ConfigStore = new_config_st()

_gemm_st_acc_i8 = (
    "gemmini_extended_config_st({dst}.strides[0]*1, {act}, {scale}[0]);\n"
    + "gemmini_extended_mvout( ((uint64_t) &{dst_data}), (uint32_t) &{src_data}, {m}, {n} );"
)


@instr(_gemm_st_acc_i8)
def st_acc_i8(
    n: size,
    m: size,
    scale: f32,
    act: bool,
    src: [i32][n, 16] @ GEMM_ACCUM,
    dst: [i8][n, m] @ DRAM,
):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 1) == 1
    assert stride(src, 0) == 16
    assert stride(src, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            src_tmp: i32
            src_tmp = src[i, j]
            tmp: f32
            acc_scale(src_tmp, tmp, scale)
            tmp2: i8
            clamp(tmp, tmp2)
            if act == True:
                tmp2 = relu(tmp2)
            dst[i, j] = tmp2


_gemm_config_st_acc_i8 = (
    "gemmini_extended_config_st({dst_stride}, {act}, {scale}[0]);\n"
)


@instr(_gemm_config_st_acc_i8)
def config_st_acc_i8(scale: f32, dst_stride: stride, act: bool):
    ConfigStore.scale = scale
    ConfigStore.dst_stride = dst_stride
    ConfigStore.act = act


_gemm_st_acc_i8 = "gemmini_extended_mvout( ((uint64_t) &{dst_data}), (uint32_t) &{src_data}, {m}, {n} );"


@instr(_gemm_st_acc_i8)
def do_st_acc_i8(
    n: size, m: size, src: [i32][n, 16] @ GEMM_ACCUM, dst: [i8][n, m] @ DRAM
):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 1) == 1
    assert stride(src, 0) == 16
    assert stride(src, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            src_tmp: i32
            src_tmp = src[i, j]
            tmp: f32
            acc_scale(src_tmp, tmp, ConfigStore.scale)
            tmp2: i8
            clamp(tmp, tmp2)
            if ConfigStore.act == True:
                tmp2 = relu(tmp2)
            dst[i, j] = tmp2


def make_st_acc_i8_v2(p=st_acc_i8):
    p = rename(st_acc_i8, "st_acc_i8_v2")
    p = bind_config(p, "scale", ConfigStore, "scale")
    p = reorder_stmts(p, "tmp : _ ; ConfigStore.scale = _")
    p = reorder_stmts(p, "src_tmp = _ ; ConfigStore.scale = _")
    p = reorder_stmts(p, "src_tmp : _ ; ConfigStore.scale = _")
    p = old_fission_after(p, "ConfigStore.scale = _", n_lifts=2)
    p = write_config(
        p,
        p.find("ConfigStore.scale = _").after(),
        ConfigStore,
        "dst_stride",
        "stride(dst, 0)",
    )
    # p = p.configwrite_after('ConfigStore.scale = _', ConfigStore, 'dst_stride', 'stride(dst, 0)')
    p = bind_config(p, "act", ConfigStore, "act")
    p = reorder_stmts(p, "clamp(_) ; ConfigStore.act = _")
    p = reorder_stmts(p, "tmp2 : _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "acc_scale(_) ; ConfigStore.act = _")
    p = reorder_stmts(p, "tmp : _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "src_tmp = _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "src_tmp : _ ; ConfigStore.act = _")
    p = old_fission_after(p, "ConfigStore.act = _", n_lifts=2)
    p = replace(p, "for i in _:_", do_st_acc_i8)
    p = replace(
        p,
        "ConfigStore.scale = _ ;" "ConfigStore.dst_stride = _ ;" "ConfigStore.act = _",
        config_st_acc_i8,
    )
    return p


st_acc_i8_v2 = make_st_acc_i8_v2()


def make_st_acc_i8_s2_v2(p=st_acc_i8):
    p = rename(p, "st_acc_i8_s2_v2")
    p = bind_config(p, "scale", ConfigStore, "scale")
    p = reorder_stmts(p, "tmp : _ ; ConfigStore.scale = _")
    p = reorder_stmts(p, "src_tmp = _ ; ConfigStore.scale = _")
    p = reorder_stmts(p, "src_tmp : _ ; ConfigStore.scale = _")
    p = old_fission_after(p, "ConfigStore.scale = _", n_lifts=2)
    p = write_config(
        p,
        p.find("ConfigStore.scale = _").after(),
        ConfigStore,
        "dst_stride",
        "stride(dst, 0)",
    )
    # p = p.configwrite_after('ConfigStore.scale = _', ConfigStore, 'dst_stride', 'stride(dst, 2)')
    p = bind_config(p, "act", ConfigStore, "act")
    p = reorder_stmts(p, "clamp(_) ; ConfigStore.act = _")
    p = reorder_stmts(p, "tmp2 : _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "acc_scale(_) ; ConfigStore.act = _")
    p = reorder_stmts(p, "tmp : _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "src_tmp = _ ; ConfigStore.act = _")
    p = reorder_stmts(p, "src_tmp : _ ; ConfigStore.act = _")
    p = old_fission_after(p, "ConfigStore.act = _", n_lifts=2)
    p = replace(p, "for i in _:_", do_st_acc_i8)
    p = replace(
        p,
        "ConfigStore.scale = _ ;" "ConfigStore.dst_stride = _ ;" "ConfigStore.act = _",
        config_st_acc_i8,
    )
    return p


st_acc_i8_s2_v2 = make_st_acc_i8_s2_v2()

_gemm_st_acc_i32 = (
    "gemmini_extended_config_st({dst}.strides[0]*4, 0, 1.0f);\n"
    + "gemmini_extended_mvout( ((uint64_t) &{dst_data}), "
    + "((uint32_t) &{src_data} | 0x20000000), {m}, {n} );"
)


@instr(_gemm_st_acc_i32)
def st_acc_i32(
    n: size, m: size, src: [i32][n, 16] @ GEMM_ACCUM, dst: [i32][n, m] @ DRAM
):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 1) == 1
    assert stride(src, 0) == 16
    assert stride(src, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = src[i, j]


_gemm_config_zero = "gemmini_extended3_config_ld(0, 1.0f, 0, 0);\n"


@instr(_gemm_config_zero)
def config_zero():
    ConfigLoad.src_stride = 0


_gemm_do_zero = "gemmini_extended_mvin( 0, ((uint64_t) &{dst_data})," + "{m}, {n} );"


@instr(_gemm_do_zero)
def do_zero_i8(
    n: size,
    m: size,
    dst: [i8][n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = 0.0


_gemm_zero = (
    "gemmini_extended3_config_ld(0, 1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( 0, ((uint64_t) &{dst_data}),"
    + "{m}, {n} );"
)


@instr(_gemm_zero)
def zero_i8(
    n: size,
    m: size,
    dst: [i8][n, 16] @ GEMM_SCRATCH,
):
    assert n <= 16
    assert m <= 16
    assert stride(dst, 0) == 16
    assert stride(dst, 1) == 1

    pass

    for i in seq(0, n):
        for j in seq(0, m):
            dst[i, j] = 0.0


def make_zero_i8_v2(p=zero_i8):
    p = rename(p, "zero_i8_v2")
    p = write_config(p, p.body().before(), ConfigLoad, "src_stride", "0")
    # p = p.configwrite_after('pass', ConfigLoad, 'src_stride', '0')
    p = replace(p, "for i in _:_", do_zero_i8)
    p = replace(p, "ConfigLoad.src_stride = _", config_zero)
    return p


zero_i8_v2 = make_zero_i8_v2()

do_zero_acc_i32 = rename(do_zero_i8, "do_zero_acc_i32")
do_zero_acc_i32 = set_prec_mem(do_zero_acc_i32, "dst", "i32", GEMM_ACCUM)
do_zero_acc_i32 = make_instr(do_zero_acc_i32, _gemm_do_zero)
zero_acc_i32 = rename(zero_i8, "zero_acc_i32")
zero_acc_i32 = set_prec_mem(zero_acc_i32, "dst", "i32", GEMM_ACCUM)
zero_acc_i32 = make_instr(zero_acc_i32, _gemm_zero)


def make_zero_acc_i32_v2(p=zero_acc_i32):
    p = rename(p, "zero_acc_i32_v2")
    p = write_config(p, p.body().before(), ConfigLoad, "src_stride", "0")
    # p = p.configwrite_after('pass', ConfigLoad, 'src_stride', '0')
    p = replace(p, "for i in _:_", do_zero_acc_i32)
    p = replace(p, "ConfigLoad.src_stride = _", config_zero)
    return p


zero_acc_i32_v2 = make_zero_acc_i32_v2()


def del_and_zero(p):
    p = delete_pass(p)
    p = make_instr(p, _gemm_zero)
    return p


zero_i8 = del_and_zero(zero_i8)
zero_i8_v2 = del_and_zero(zero_i8_v2)
zero_acc_i32 = del_and_zero(zero_acc_i32)
zero_acc_i32_v2 = del_and_zero(zero_acc_i32_v2)


_gemm_zero_vec = (
    "gemmini_extended3_config_ld(0, 1.0f, 0, 0);\n"
    + "gemmini_extended_mvin( 0, ((uint64_t) &{dst_data}),"
    + "16, 1 );"
)


@instr(_gemm_zero_vec)
def zero_i8_vector(
    dst: [i8][16] @ GEMM_SCRATCH,
):
    assert stride(dst, 0) == 16
    pass

    for i in seq(0, 16):
        dst[i] = 0.0


_do_gemm_zero_vec = "gemmini_extended_mvin( 0, ((uint64_t) &{dst_data})," + "16, 1 );"


@instr(_do_gemm_zero_vec)
def do_zero_i8_vector(
    dst: [i8][16] @ GEMM_SCRATCH,
):
    assert stride(dst, 0) == 16

    for i in seq(0, 16):
        dst[i] = 0.0


def make_zero_i8_vector_v2(p=zero_i8_vector):
    p = rename(p, "zero_i8_vector_v2")
    p = write_config(p, p.body().before(), ConfigLoad, "src_stride", "0")
    # p = p.configwrite_after('pass', ConfigLoad, 'src_stride', '0')
    p = replace(p, "for i in _:_", do_zero_i8_vector)
    p = replace(p, "ConfigLoad.src_stride = _", config_zero)
    p = delete_pass(p)
    p = make_instr(p, _gemm_zero_vec)
    return p


zero_i8_vector_v2 = make_zero_i8_vector_v2()

zero_i8_vector = delete_pass(zero_i8_vector)
zero_i8_vector = make_instr(zero_i8_vector, _gemm_zero_vec)


def new_config_matmul():
    @config
    class ConfigMatmul:
        done: bool

    return ConfigMatmul


ConfigMatmul = new_config_matmul()

_gemm_config_matmul = "gemmini_extended_config_ex(WS, 0, 0, 1, 0, 0);\n"


@instr(_gemm_config_matmul)
def config_matmul():
    ConfigMatmul.done = True


_gemm_matmul = (
    "gemmini_extended_preload("
    + "(uint32_t)(&{B_data}), (uint32_t)(&{C_data}), "
    + "{M}, {K}, "
    + "{M}, {N}"
    + ");\n"
    + "gemmini_extended_compute_preloaded("
    + "(uint32_t)(&{A_data}), ~((uint32_t)0), "
    + "{K}, {N}, "
    + "16, 16"
    + ");"
)


@instr(_gemm_config_matmul + _gemm_matmul)
def matmul_i8(
    N: size,
    M: size,
    K: size,
    A: [i8][N, 16] @ GEMM_SCRATCH,
    B: [i8][K, 16] @ GEMM_SCRATCH,
    C: [i32][N, 16] @ GEMM_ACCUM,
):
    assert N <= 16
    assert M <= 16
    assert K <= 16

    pass
    for i in seq(0, N):
        for j in seq(0, M):
            C[i, j] = 0.0
            for k in seq(0, K):
                a: i32
                b: i32

                a = A[i, k]
                b = B[k, j]

                C[i, j] += a * b


@instr(_gemm_matmul)
def do_matmul_i8(
    N: size,
    M: size,
    K: size,
    A: [i8][N, 16] @ GEMM_SCRATCH,
    B: [i8][K, 16] @ GEMM_SCRATCH,
    C: [i32][N, 16] @ GEMM_ACCUM,
):
    assert N <= 16
    assert M <= 16
    assert K <= 16

    for i in seq(0, N):
        for j in seq(0, M):
            C[i, j] = 0.0
            for k in seq(0, K):
                a: i32
                b: i32

                a = A[i, k]
                b = B[k, j]

                C[i, j] += a * b


def make_matmul_i8_v2(p=matmul_i8):
    p = rename(p, "matmul_i8_v2")
    p = write_config(p, p.body().before(), ConfigMatmul, "done", "True")
    # p = p.configwrite_after('pass', ConfigMatmul, 'done', 'True')
    p = replace(p, "for i in _:_", do_matmul_i8)
    p = replace(p, "ConfigMatmul.done = True", config_matmul)
    p = delete_pass(p)
    p = make_instr(p, _gemm_matmul)
    return p


matmul_i8_v2 = make_matmul_i8_v2()

matmul_i8 = delete_pass(matmul_i8)
matmul_i8 = make_instr(matmul_i8, _gemm_config_matmul + _gemm_matmul)


_gemm_matmul_acc = (
    "gemmini_extended_preload("
    + "(uint32_t)(&{B_data}), (uint32_t)(&{C_data}) | 0x40000000, "
    + "{M}, {K}, "
    + "{M}, {N}"
    + ");\n"
    + "gemmini_extended_compute_preloaded("
    + "(uint32_t)(&{A_data}), ~((uint32_t)0), "
    + "{K}, {N}, "
    + "16, 16"
    + ");"
)


@instr(_gemm_matmul_acc)
def matmul_acc_i8(
    N: size,
    M: size,
    K: size,
    A: [i8][N, 16] @ GEMM_SCRATCH,
    B: [i8][K, 16] @ GEMM_SCRATCH,
    C: [i32][N, 16] @ GEMM_ACCUM,
):
    assert N <= 16
    assert M <= 16
    assert K <= 16

    pass
    for i in seq(0, N):
        for j in seq(0, M):
            for k in seq(0, K):
                a: i32
                b: i32

                a = A[i, k]
                b = B[k, j]

                C[i, j] += a * b


@instr(_gemm_matmul_acc)
def do_matmul_acc_i8(
    N: size,
    M: size,
    K: size,
    A: [i8][N, 16] @ GEMM_SCRATCH,
    B: [i8][K, 16] @ GEMM_SCRATCH,
    C: [i32][N, 16] @ GEMM_ACCUM,
):
    assert N <= 16
    assert M <= 16
    assert K <= 16

    for i in seq(0, N):
        for j in seq(0, M):
            for k in seq(0, K):
                a: i32
                b: i32

                a = A[i, k]
                b = B[k, j]

                C[i, j] += a * b


def make_matmul_acc_i8_v2(p=matmul_acc_i8):
    p = rename(p, "matmul_acc_i8_v2")
    p = write_config(p, p.body().before(), ConfigMatmul, "done", "True")
    # p = p.configwrite_after('pass', ConfigMatmul, 'done', 'True')
    p = replace(p, "for i in _:_", do_matmul_acc_i8)
    p = replace(p, "ConfigMatmul.done = True", config_matmul)
    p = delete_pass(p)
    p = make_instr(p, _gemm_matmul_acc)
    return p


matmul_acc_i8_v2 = make_matmul_acc_i8_v2()

matmul_acc_i8 = delete_pass(matmul_acc_i8)
matmul_acc_i8 = make_instr(matmul_acc_i8, _gemm_config_matmul + _gemm_matmul_acc)

# --------------------------------------------------------------------------- #
#
# --------------------------------------------------------------------------- #



from .inspection import *

exo_exceptions_ = {
    ValueError,
    TypeError,
    SchedulingError,
    InvalidCursorError,
}


def attempt(op, errs=exo_exceptions_):
    errs = tuple(errs)

    def rewrite(p, *args, rs=False, **kwargs):
        try:
            res = op(p, *args, **kwargs), True
        except errs:
            res = p, False
        if not rs:
            res = res[0]
        return res

    return rewrite


def apply(op):
    def rewrite(proc, cursors, *args, **kwargs):
        for c in cursors:
            proc = op(proc, c, *args, **kwargs)
        return proc

    return rewrite


def predicate(op, pred):
    def rewrite(proc, *args, **kwargs):
        if pred(proc, *args, **kwargs):
            proc = op(proc, *args, **kwargs)
        return proc

    return rewrite


def make_pass(op, trav_start=nlr_stmts):
    def rewrite(proc, block=InvalidCursor(), *args, **kwargs):
        stmts = trav_start(proc, block)
        return apply(op)(proc, stmts, *args, **kwargs)

    return rewrite


def lift_rc(op, attr):
    def rewrite(*args, **kwargs):
        proc, cursors = op(*args, **kwargs, rc=True)
        c = getattr(cursors, attr)
        return proc, c

    return rewrite


def repeate(op):
    op_attempt = attempt(op)

    def rewrite(p, *args, **kwargs):
        success = True
        while success:
            p, success = op_attempt(p, *args, **kwargs, rs=True)
        return p

    return rewrite


__all__ = ["apply", "attempt", "make_pass", "lift_rc", "repeate", "predicate"]



import argparse
import importlib
import importlib.machinery
import importlib.util
import inspect
import sys

sys.setrecursionlimit(10000)

from pathlib import Path

import exo


def main():
    parser = argparse.ArgumentParser(
        prog=Path(sys.argv[0]).name, description="Compile an Exo library."
    )
    parser.add_argument(
        "-o",
        "--outdir",
        metavar="OUTDIR",
        required=True,
        help="output directory for build artifacts",
    )
    parser.add_argument("--stem", required=True, help="base name for .c and .h files")
    parser.add_argument("source", type=str, nargs="+", help="source file to compile")
    parser.add_argument(
        "--version",
        action="version",
        version=f"%(prog)s version {exo.__version__}",
        help="print the version and exit",
    )

    args = parser.parse_args()

    outdir = Path(args.outdir)
    outdir.mkdir(parents=True, exist_ok=True)

    library = [
        proc
        for mod in args.source
        for proc in get_procs_from_module(load_user_code(mod))
    ]

    exo.compile_procs(library, outdir, f"{args.stem}.c", f"{args.stem}.h")
    write_depfile(outdir, args.stem)


def write_depfile(outdir, stem):
    modules = set()
    for mod in sys.modules.values():
        try:
            modules.add(inspect.getfile(mod))
        except TypeError:
            pass  # this is the case for built-in modules

    c_file = outdir / f"{stem}.c"
    h_file = outdir / f"{stem}.h"
    depfile = outdir / f"{stem}.d"

    sep = " \\\n  "
    deps = sep.join(sorted(modules))
    contents = f"{c_file} {h_file} : {deps}"

    depfile.write_text(contents)


def get_procs_from_module(user_module):
    symbols = dir(user_module)
    has_export_list = "__all__" in symbols
    if has_export_list:
        exported_symbols = user_module.__dict__["__all__"]
    else:
        exported_symbols = symbols
    library = []
    for sym in exported_symbols:
        if not sym.startswith("_"):
            fn = getattr(user_module, sym)
            if isinstance(fn, exo.Procedure) and not fn.is_instr():
                library.append(fn)
    return library


def load_user_code(path):
    module_path = Path(path).resolve(strict=True)
    module_name = module_path.stem
    module_path = str(module_path)
    loader = importlib.machinery.SourceFileLoader(module_name, module_path)
    spec = importlib.util.spec_from_loader(module_name, loader)
    user_module = importlib.util.module_from_spec(spec)
    loader.exec_module(user_module)
    return user_module


if __name__ == "__main__":
    main()



from functools import wraps as _wraps


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Expose the built-in Scheduling operators here

from ..API import SchedulingError, Procedure

from ..API_scheduling import (
    is_atomic_scheduling_op,
    # argument processing
    sched_op,
    BlockCursorA,
    ProcA,
    BoolA,
    ArgOrAllocCursorA,
    FormattedExprStr,
    # basic operations
    simplify,
    rename,
    make_instr,
    #
    # general statement and expression operations
    insert_pass,
    delete_pass,
    reorder_stmts,
    rewrite_expr,
    bind_expr,
    commute_expr,
    left_reassociate_expr,
    #
    # subprocedure oriented operations
    extract_subproc,
    inline,
    replace,
    call_eqv,
    #
    # precision, memory, and window annotation setting
    set_precision,
    set_window,
    set_memory,
    #
    # Configuration modifying operations
    bind_config,
    delete_config,
    write_config,
    #
    # buffer and window oriented operations
    expand_dim,
    resize_dim,
    rearrange_dim,
    divide_dim,
    mult_dim,
    sink_alloc,
    lift_alloc,
    delete_buffer,
    reuse_buffer,
    inline_window,
    stage_mem,
    unroll_buffer,
    #
    # loop rewriting
    parallelize_loop,
    divide_with_recompute,
    divide_loop,
    mult_loops,
    cut_loop,
    join_loops,
    shift_loop,
    reorder_loops,
    merge_writes,
    split_write,
    fold_into_reduce,
    inline_assign,
    lift_reduce_constant,
    fission,
    fuse,
    remove_loop,
    add_loop,
    unroll_loop,
    #
    # guard rewriting
    lift_scope,
    eliminate_dead_code,
    specialize,
    #
    # deprecated scheduling operations
    add_unsafe_guard,
    #
    # to be replaced by stdlib compositions eventually
    autofission,
    autolift_alloc,
)

from .analysis import check_call_mem_types
from ..API_cursors import *
from ..LoopIR_unification import UnificationError as _UnificationError


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Higher-order Scheduling operations


def repeat(sched, n_times=None, verbose=False):
    """
    TODO: Documentation
    """
    if n_times is not None and (not isinstance(n_times, int) or n_times < 1):
        raise TypeError("expected n_times to be None or a positive int")

    @_wraps(sched)
    def repeated_sched(proc, *args, **kwargs):
        def do_iter():
            nonlocal proc
            local_args = args.copy() if isinstance(args, list) else args
            local_kwargs = kwargs.copy()
            proc = sched(proc, *local_args, **local_kwargs)

        if n_times is None:
            try:
                while True:
                    do_iter()
            except (SchedulingError, TypeError, ValueError) as err:
                if verbose:
                    print("repeat ended with error", err)
        else:
            for i in range(n_times):
                do_iter()

        return proc

    repeated_sched.__name__ = f"repeat_{sched.__name__}"
    return repeated_sched


_sched_seq_err = """
sched_seq(proc, sched_list) expects an iterable object `sched_list`
containing a sequence of scheduling operations to apply to `proc`.
Each scheduling operation must either be a Python Callable object
(such as a function) which will be passed one argument---the `proc`---or
the scheduling operation must be a list/tuple whose first entry is
a scheduling function/operation and whose remaining entries are arguments
to be passed to that scheduling function.

e.g.
proc = sched_seq(proc,[
    (rename, 'foo'),
    (set_memory, 'A', SPECIAL_DRAM),
    ...
])
"""


def sched_seq(proc, sched_list):
    for s in sched_list:
        if callable(s):
            proc = s(proc)
        elif isinstance(s, (list, tuple)):
            if len(s) == 0:
                raise TypeError(
                    _sched_seq_err
                    + "\n"
                    + "expected scheduling operation list/tuple "
                    + "to have at least one entry"
                )
            s_call = s[0]
            s_args = s[1:]
            proc = s_call(proc, *s_args)
        else:
            raise TypeError(_sched_seq_err)
    return proc


def loop_hack(sched, find_func, verbose=False):
    """
    DEPRECATED

    This higher level method is a patch to help migrate code to the
    new interface, but before the cursor forwarding implementation is
    complete.  It should be eliminated and replaced with a cursor-friendly
    version once forwarding is finished being implemented
    """

    @_wraps(sched)
    def loop_hack_sched(proc, *args, **kwargs):
        match_len = len(find_func(proc))
        for i in range(0, match_len):
            cursor = find_func(proc)[i]
            proc = sched(proc, cursor, *args, **kwargs)

        return proc

    loop_hack_sched.__name__ = f"loop_hack_{sched.__name__}"
    return loop_hack_sched


class MemoryError(Exception):
    def __init__(self, msg):
        self._err_msg = str(msg)

    def __str__(self):
        return self._err_msg


@sched_op([BlockCursorA, ProcA, BoolA])
def call_site_mem_aware_replace(proc, block_cursor, subproc, quiet=False):
    proc = replace(proc, block_cursor, subproc, quiet=quiet)

    def check_all_calls(body_cursor):
        check_passed = True
        for cursor in body_cursor:
            if isinstance(cursor, CallCursor):
                check_passed = check_passed and check_call_mem_types(cursor)
            elif isinstance(cursor, IfCursor):
                check_passed = check_passed and check_all_calls(cursor.body())
                if type(cursor.orelse()) is not InvalidCursor:
                    check_passed = check_passed and check_all_calls(cursor.orelse())
            elif isinstance(cursor, ForCursor):
                check_passed = check_passed and check_all_calls(cursor.body())
        return check_passed

    if not check_all_calls(proc.body()):
        raise MemoryError(
            "replace failed due to memory type mismatch between block and subproc"
        )

    return proc


def _replace_helper(proc, subprocs, mem_aware, once):

    if not isinstance(subprocs, list):
        subprocs = [subprocs]

    for subproc in subprocs:
        assert isinstance(subproc, Procedure), "expected Procedure as 2nd argument"

    patterns = {
        AssignCursor: "_ = _",
        ReduceCursor: "_ += _",
        AssignConfigCursor: "TODO",
        PassCursor: "TODO",
        IfCursor: "TODO",
        ForCursor: "for _ in _: _",
        AllocCursor: "TODO",
        CallCursor: "TODO",
        WindowStmtCursor: "TODO",
    }

    for subproc in subprocs:
        body = subproc.body()
        pattern = patterns[type(body[0])]
        i = 0
        while True:
            try:
                block = proc.find(f"{pattern} #{i}").expand(0, len(body) - 1)
                if len(block) != len(body):
                    raise _UnificationError("Unification failed due to length mismatch")

                if mem_aware:
                    proc = call_site_mem_aware_replace(proc, block, subproc, quiet=True)
                else:
                    proc = replace(proc, block, subproc, quiet=True)
                if once:
                    break
            except (TypeError, SchedulingError) as e:
                if "failed to find matches" in str(e):
                    break
                raise
            except (
                _UnificationError,
                MemoryError,
                NotImplementedError,
            ):
                i += 1

    return proc


def replace_all(proc, subprocs, mem_aware=True):
    """
    Givin a proc and subprocs, replace the body of proc with subproc
    as much as possible.

    args:
        subprocs  - list of subprocedures (or instruction definitions) to
                    be replaced with
        mem_aware - if True, replace will only suceed when memory annotation
                    also matches
    """

    return _replace_helper(proc, subprocs, mem_aware, False)


def replace_once(proc, subprocs, mem_aware=True):
    """
    Givin a proc and subprocs, replace the body of proc with subproc only once.

    args:
        subprocs  - list of subprocedures (or instruction definitions) to
                    be replaced with
        mem_aware - if True, replace will only suceed when memory annotation
                    also matches
    """

    return _replace_helper(proc, subprocs, mem_aware, True)


def lift_if(proc, cursor, n_lifts=1):
    """
    Move the indicated If-statement upwards through other control-flow
    for a total of n_lifts times.

    args:
        cursor       - cursor to the innermost if statement to lift up
        n_lifts      - number of times to lift the if statement up

    rewrite: (one example)
        `for i in _:`
        `    if p:`
        `        s1`
        `    else:`
        `        s2`
        ->
        `if p:`
        `    for i in _:`
        `        s1`
        `else:`
        `    for i in _:`
        `        s2`
    """
    orig_proc = proc
    for i in range(n_lifts):
        try:
            proc = lift_scope(proc, cursor)
        except SchedulingError as e:
            raise SchedulingError(
                f"Could not fully lift if statement! {n_lifts-i} lift(s) remain! {str(e)}",
                orig=orig_proc,
                proc=proc,
            ) from e
    return proc



from .LoopIR import LoopIR, T, Effects
from .prelude import *

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Printing Functions

op_prec = {
    "ternary": 5,
    #
    "or": 10,
    #
    "and": 20,
    #
    "<": 30,
    ">": 30,
    "<=": 30,
    ">=": 30,
    "==": 30,
    #
    "+": 40,
    "-": 40,
    #
    "*": 50,
    "/": 50,
    "%": 50,
    #
    # unary - 60
}


@extclass(Effects.Var)
@extclass(Effects.Not)
@extclass(Effects.Const)
@extclass(Effects.BinOp)
@extclass(Effects.Stride)
@extclass(Effects.Select)
def __str__(self):
    return _exprstr(self)


del __str__


def _exprstr(e, prec=0):
    if isinstance(e, Effects.Var):
        return str(e.name)
    elif isinstance(e, Effects.Not):
        return f"(not {e.arg})"
    elif isinstance(e, Effects.Const):
        return str(e.val)
    elif isinstance(e, Effects.BinOp):
        local_prec = op_prec[e.op]
        lhs = _exprstr(e.lhs, prec=local_prec)
        rhs = _exprstr(e.rhs, prec=local_prec + 1)
        if local_prec < prec:
            return f"({lhs} {e.op} {rhs})"
        else:
            return f"{lhs} {e.op} {rhs}"
    elif isinstance(e, Effects.Stride):
        return f"stride({e.name},{e.dim})"
    elif isinstance(e, Effects.Select):
        local_prec = op_prec["ternary"]
        cond = _exprstr(e.cond)
        tcase = _exprstr(e.tcase, prec=local_prec + 1)
        fcase = _exprstr(e.fcase, prec=local_prec + 1)
        if local_prec < prec:
            return f"(({cond})? {tcase} : {fcase})"
        else:
            return f"({cond})? {tcase} : {fcase}"
    elif isinstance(e, Effects.ConfigField):
        return f"{e.config.name()}.{e.field}"
    else:
        assert False, "bad case"


@extclass(Effects.effect)
def __str__(self):
    return _effect_as_str(self)


del __str__


def _effect_as_str(e):
    assert isinstance(e, Effects.effect)

    def name(sym):
        return str(sym)

    def esstr(es, tab="  "):
        lines = []
        buf = name(es.buffer)
        loc = "(" + ",".join([str(l) for l in es.loc]) + ")"
        if len(es.names) == 0:
            names = ""
        else:
            names = f"for ({','.join([name(n) for n in es.names])}) in Z"

        if es.pred is None:
            lines.append(f"{tab}{{ {buf} : {loc} {names} }}")
        else:
            lines.append(f"{tab}{{ {buf} : {loc} {names} if")
            tab += "  "
            pred = str(es.pred)
            lines.append(f"{tab}{pred} }}")

        return "\n".join(lines)

    def cestr(ce, tab="  "):
        val, pred = "", ""
        if ce.value:
            val = f" = {ce.value}"
        if ce.pred:
            pred = f" if {ce.pred}"
        return f"{ce.config.name()}.{ce.field}{val}{pred}"

    eff_str = ""
    if len(e.reads) > 0:
        eff_str += "Reads:\n"
        eff_str += "\n".join([esstr(es) for es in e.reads])
        eff_str += "\n"
    if len(e.writes) > 0:
        eff_str += f"Writes:\n  "
        eff_str += "\n".join([esstr(es) for es in e.writes])
        eff_str += "\n"
    if len(e.reduces) > 0:
        eff_str += f"Reduces:\n  "
        eff_str += "\n".join([esstr(es) for es in e.reduces])
        eff_str += "\n"
    if len(e.config_reads) > 0:
        eff_str += f"Config Reads:\n"
        eff_str += "\n".join([cestr(ce) for ce in e.config_reads])
        eff_str += "\n"
    if len(e.config_writes) > 0:
        eff_str += f"Config Writes:\n"
        eff_str += "\n".join([cestr(ce) for ce in e.config_writes])
        eff_str += "\n"

    return eff_str


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Substitution of Effect Variables in an effect


@extclass(Effects.expr)
def negate(self):
    return negate_expr(self)


del negate


def negate_expr(e):
    Tbool = T.bool
    assert e.type == Tbool, "can only negate predicates"
    if isinstance(e, Effects.Const):
        return Effects.Const(not e.val, e.type, e.srcinfo)
    elif isinstance(e, Effects.Var) or isinstance(e, Effects.ConfigField):
        return Effects.Not(e, e.type, e.srcinfo)
    elif isinstance(e, Effects.Not):
        return e.arg
    elif isinstance(e, Effects.BinOp):

        def change_op(op, lhs=e.lhs, rhs=e.rhs):
            return Effects.BinOp(op, lhs, rhs, e.type, e.srcinfo)

        if e.op == "and":
            return change_op("or", negate_expr(e.lhs), negate_expr(e.rhs))
        elif e.op == "or":
            return change_op("and", negate_expr(e.lhs), negate_expr(e.rhs))
        elif e.op == ">":
            return change_op("<=")
        elif e.op == "<":
            return change_op(">=")
        elif e.op == ">=":
            return change_op("<")
        elif e.op == "<=":
            return change_op(">")
        elif e.op == "==":
            if e.lhs.type is Tbool and e.rhs.type is Tbool:
                l = Effects.BinOp("and", e.lhs, negate_expr(e.rhs), Tbool, e.srcinfo)
                r = Effects.BinOp("and", negate_expr(e.lhs), e.rhs, Tbool, e.srcinfo)

                return Effects.BinOp("or", l, r, Tbool, e.srcinfo)
            elif e.lhs.type.is_indexable() and e.rhs.type.is_indexable():
                return Effects.BinOp(
                    "or", change_op("<"), change_op(">"), Tbool, e.srcinfo
                )
            else:
                assert False, "TODO: add != support explicitly..."
    elif isinstance(e, Effects.Select):
        return Effects.Select(
            e.cond, negate_expr(e.tcase), negate_expr(e.fcase), e.type, e.srcinfo
        )
    assert False, "bad case"


@extclass(Effects.effect)
@extclass(Effects.effset)
@extclass(Effects.expr)
def subst(self, env):
    return eff_subst(env, self)


del subst


def eff_subst(env, eff):
    if isinstance(eff, Effects.effset):
        assert all(nm not in env for nm in eff.names)
        buf = env[eff.buffer] if eff.buffer in env else eff.buffer
        pred = eff_subst(env, eff.pred) if eff.pred else None
        return Effects.effset(
            buf, [eff_subst(env, e) for e in eff.loc], eff.names, pred, eff.srcinfo
        )
    elif isinstance(eff, Effects.config_eff):
        value = eff_subst(env, eff.value) if eff.value else None
        pred = eff_subst(env, eff.pred) if eff.pred else None
        return Effects.config_eff(eff.config, eff.field, value, pred, eff.srcinfo)
    elif isinstance(eff, Effects.Var):
        return env[eff.name] if eff.name in env else eff
    elif isinstance(eff, Effects.Not):
        return Effects.Not(eff_subst(env, eff.arg), eff.type, eff.srcinfo)
    elif isinstance(eff, Effects.Const):
        return eff
    elif isinstance(eff, Effects.BinOp):
        return Effects.BinOp(
            eff.op,
            eff_subst(env, eff.lhs),
            eff_subst(env, eff.rhs),
            eff.type,
            eff.srcinfo,
        )
    elif isinstance(eff, Effects.Stride):
        name = env[eff.name] if eff.name in env else eff
        return Effects.Stride(name, eff.dim, eff.type, eff.srcinfo)
    elif isinstance(eff, Effects.Select):
        return Effects.Select(
            eff_subst(env, eff.cond),
            eff_subst(env, eff.tcase),
            eff_subst(env, eff.fcase),
            eff.type,
            eff.srcinfo,
        )
    elif isinstance(eff, Effects.ConfigField):
        return eff
    elif isinstance(eff, Effects.effect):
        return Effects.effect(
            [eff_subst(env, es) for es in eff.reads],
            [eff_subst(env, es) for es in eff.writes],
            [eff_subst(env, es) for es in eff.reduces],
            [eff_subst(env, ce) for ce in eff.config_reads],
            [eff_subst(env, ce) for ce in eff.config_writes],
            eff.srcinfo,
        )
    else:
        assert False, f"bad case: {type(eff)}"


@extclass(Effects.effect)
@extclass(Effects.effset)
@extclass(Effects.expr)
def config_subst(self, env):
    return _subcfg(env, self)


del config_subst


def _subcfg(env, eff):
    if isinstance(eff, Effects.effset):
        return Effects.effset(
            eff.buffer,
            [_subcfg(env, e) for e in eff.loc],
            eff.names,
            _subcfg(env, eff.pred) if eff.pred else None,
            eff.srcinfo,
        )
    elif isinstance(eff, Effects.config_eff):
        value = _subcfg(env, eff.value) if eff.value else None
        pred = _subcfg(env, eff.pred) if eff.pred else None
        return Effects.config_eff(eff.config, eff.field, value, pred, eff.srcinfo)
    elif isinstance(eff, (Effects.Var, Effects.Const)):
        return eff
    elif isinstance(eff, Effects.Not):
        return Effects.Not(_subcfg(env, eff.arg), eff.type, eff.srcinfo)
    elif isinstance(eff, Effects.BinOp):
        return Effects.BinOp(
            eff.op, _subcfg(env, eff.lhs), _subcfg(env, eff.rhs), eff.type, eff.srcinfo
        )
    elif isinstance(eff, Effects.Stride):
        return eff
    elif isinstance(eff, Effects.Select):
        return Effects.Select(
            _subcfg(env, eff.cond),
            _subcfg(env, eff.tcase),
            _subcfg(env, eff.fcase),
            eff.type,
            eff.srcinfo,
        )
    elif isinstance(eff, Effects.ConfigField):
        if (eff.config, eff.field) in env:
            return env[(eff.config, eff.field)]
        else:
            return eff
    elif isinstance(eff, Effects.effect):
        return Effects.effect(
            [_subcfg(env, es) for es in eff.reads],
            [_subcfg(env, es) for es in eff.writes],
            [_subcfg(env, es) for es in eff.reduces],
            [_subcfg(env, ce) for ce in eff.config_reads],
            [_subcfg(env, ce) for ce in eff.config_writes],
            eff.srcinfo,
        )
    else:
        assert False, f"bad case: {type(eff)}"


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Querying of the effect of a block of already annotated statements


def get_effect_of_stmts(body):
    assert len(body) > 0
    eff = eff_null(body[0].srcinfo)
    for s in reversed(body):
        if isinstance(s, LoopIR.Alloc):
            eff = eff_remove_buf(s.name, eff)
        elif s.eff is not None:
            eff = eff_concat(s.eff, eff)
    return eff


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Construction/Composition Functions


def eff_null(srcinfo=null_srcinfo()):
    return Effects.effect([], [], [], [], [], srcinfo)


def eff_read(buf, loc, srcinfo=null_srcinfo()):
    read = Effects.effset(buf, loc, [], None, srcinfo)
    return Effects.effect([read], [], [], [], [], srcinfo)


def eff_write(buf, loc, srcinfo=null_srcinfo()):
    write = Effects.effset(buf, loc, [], None, srcinfo)
    return Effects.effect([], [write], [], [], [], srcinfo)


def eff_reduce(buf, loc, srcinfo=null_srcinfo()):
    reduction = Effects.effset(buf, loc, [], None, srcinfo)
    return Effects.effect([], [], [reduction], [], [], srcinfo)


def eff_config_read(config, field, srcinfo=null_srcinfo()):
    read = Effects.config_eff(config, field, None, None, srcinfo)
    return Effects.effect([], [], [], [read], [], srcinfo)


def eff_config_write(config, field, value, srcinfo=null_srcinfo()):
    write = Effects.config_eff(config, field, value, None, srcinfo)
    return Effects.effect([], [], [], [], [write], srcinfo)


def _and_preds(a, b):
    return (
        a
        if b is None
        else b
        if a is None
        else Effects.BinOp("and", a, b, T.bool, a.srcinfo)
    )


def _or_preds(a, b):
    return (
        None if a is None or b is None else Effects.BinOp("or", a, b, T.bool, a.srcinfo)
    )


def eff_union(e1, e2, srcinfo=None):
    srcinfo = srcinfo or e1.srcinfo

    return Effects.effect(
        e1.reads + e2.reads,
        e1.writes + e2.writes,
        e1.reduces + e2.reduces,
        e1.config_reads + e2.config_reads,
        e1.config_writes + e2.config_writes,
        srcinfo,
    )


# handle complex logic for Configuration when computing
#   EffectOf( s1 ; s2 ) in terms of EffectOf( s1 ) and EffectOf( s2 )
def eff_concat(e1, e2, srcinfo=None):
    srcinfo = srcinfo or e1.srcinfo

    # Step 1: substitute references in e2 to fields written by e1
    # value expression of config.field after this effect
    def write_val(ce):
        assert ce.value is not None
        if not ce.pred:
            return ce.value
        else:
            # TODO: Fix! I'm not sure what is the intent here..
            old_val = Effects.ConfigField(ce.config, ce.field, T.bool, srcinfo)
            return Effects.Select(ce.pred, ce.value, old_val, T.bool, srcinfo)

    # substitute on the basis of writes in the first effect
    env = {(ce.config, ce.field): write_val(ce) for ce in e1.config_writes}
    e2 = e2.config_subst(env)

    # Step 2: merge writes from the two effects to the same field
    def merge_writes(config_writes_1, config_writes_2):
        cws1 = {(w.config, w.field): w for w in config_writes_1}
        cws2 = {(w.config, w.field): w for w in config_writes_2}
        overlap = set(cws1.keys()).intersection(set(cws2.keys()))

        def merge(w1, w2):
            # in the case of the second write being unconditional
            if w2.pred is None:
                return w2
            else:
                typ = w1.config.lookup(w1.field)[1]
                assert typ == w2.config.lookup(w2.field)[1]

                pred = _or_preds(w1.pred, w2.pred)
                val = Effects.Select(w2.pred, w2.value, w1.value, typ, w2.srcinfo)

                return Effects.config_eff(w1.config, w1.field, val, pred, w2.srcinfo)

        return (
            [cws1[w] for w in cws1 if w not in overlap]
            + [cws2[w] for w in cws2 if w not in overlap]
            + [merge(cws1[key], cws2[key]) for key in overlap]
        )

    # Step 3: filter out config reads in e2 if they are just
    #         reading the config value written in e1
    def shadow_config_reads(config_writes, config_reads):
        results = []
        for read in config_reads:
            assert read.value is None

            # find the corresponding write
            write = None
            for w in config_writes:
                if w.config == read.config and w.field == read.field:
                    write = w
                    break

            # handle the shadowing
            if write is None:
                results.append(read)
            elif write.pred is None:
                # unconditional write, so remove the read
                pass
            else:
                # conditional write, so guard the read
                pred = _and_preds(read.pred, write.pred.negate())
                results.append(
                    Effects.config_eff(
                        read.config, read.field, None, pred, read.srcinfo
                    )
                )

        return results

    def shadow_reads(writes, reads):
        def shadow_read_by_write(write, read):
            def boolop(op, lhs, rhs):
                return Effects.BinOp(op, lhs, rhs, T.bool, write.srcinfo)

            loc = []
            for l1, l2 in zip(write.loc, read.loc):
                loc += [boolop("==", l1, l2).negate()]

            # construct loc predicate
            if loc == []:
                loc_e = Effects.Const(False, T.bool, write.srcinfo)
            else:
                loc_e = loc[0]
            for l in loc[1:]:
                loc_e = boolop("or", loc_e, l)

            # loc_e /\ not write.pred
            if write.pred is None:
                pred = loc_e
            else:
                pred = boolop("or", loc_e, write.pred.negate())

            pred = _and_preds(read.pred, pred)

            read = Effects.effset(
                read.buffer, read.loc, read.names + write.names, pred, read.srcinfo
            )

            return read

        def shadow_read_by_writes(writes, read):
            for w in writes:
                # find that the write
                if w.buffer == read.buffer:
                    read = shadow_read_by_write(w, read)

            return read

        return [shadow_read_by_writes(writes, r) for r in reads]

    config_reads = e1.config_reads + shadow_config_reads(
        e1.config_writes, e2.config_reads
    )
    config_writes = merge_writes(e1.config_writes, e2.config_writes)

    # TODO: Fix shadow_reads by introducing Exists
    # reads           = (e1.reads + shadow_reads(e1.writes, e2.reads))
    reads = e1.reads + e2.reads

    return Effects.effect(
        reads,
        e1.writes + e2.writes,
        e1.reduces + e2.reduces,
        config_reads,
        config_writes,
        srcinfo,
    )


def eff_remove_buf(buf, e):
    return Effects.effect(
        [es for es in e.reads if es.buffer != buf],
        [es for es in e.writes if es.buffer != buf],
        [es for es in e.reduces if es.buffer != buf],
        e.config_reads,
        e.config_writes,
        e.srcinfo,
    )


# handle conditional
def eff_filter(pred, e):
    def filter_es(es):
        return Effects.effset(
            es.buffer, es.loc, es.names, _and_preds(pred, es.pred), es.srcinfo
        )

    def filter_ce(ce):
        return Effects.config_eff(
            ce.config, ce.field, ce.value, _and_preds(pred, ce.pred), ce.srcinfo
        )

    return Effects.effect(
        [filter_es(es) for es in e.reads],
        [filter_es(es) for es in e.writes],
        [filter_es(es) for es in e.reduces],
        [filter_ce(ce) for ce in e.config_reads],
        [filter_ce(ce) for ce in e.config_writes],
        e.srcinfo,
    )


# handle for loop
def eff_bind(bind_name, e, pred=None, config_pred=None):
    assert isinstance(bind_name, Sym)

    def bind_es(es):
        return Effects.effset(
            es.buffer,
            es.loc,
            [bind_name] + es.names,
            _and_preds(pred, es.pred),
            es.srcinfo,
        )

    def filter_ce(ce):
        return Effects.config_eff(
            ce.config, ce.field, ce.value, _and_preds(pred, config_pred), ce.srcinfo
        )

    return Effects.effect(
        [bind_es(es) for es in e.reads],
        [bind_es(es) for es in e.writes],
        [bind_es(es) for es in e.reduces],
        [filter_ce(ce) for ce in e.config_reads],
        [filter_ce(ce) for ce in e.config_writes],
        e.srcinfo,
    )


"""
# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Free Variable check

@extclass(Effects.effect)
@extclass(Effects.effset)
@extclass(Effects.config_eff)
@extclass(Effects.expr)
def has_FV(self, x):
    return _is_FV(x, self)
del has_FV

def _is_FV(x, eff):
    if isinstance(eff, Effects.effset):
        if x in eff.names:
            return False
        return ( any( _is_FV(x, e) for e in eff.loc ) or
                 (eff.pred is not None and is_FV(x, eff.pred)) )
    elif isinstance(eff, Effects.config_eff):
        return ( (eff.value is not None and _is_FV(x, eff.value)) or
                 (eff.pred  is not None and _is_FV(x, eff.pred)) )
    elif isinstance(eff, Effects.Var):
        return x == eff.name
    elif isinstance(eff, Effects.Const):
        return False
    elif isinstance(eff, Effects.Not):
        return _is_FV(x, eff.arg)
    elif isinstance(eff, Effects.BinOp):
        return _is_FV(x, eff.lhs) or _is_FV(x, eff.rhs)
    elif isinstance(eff, Effects.Stride):
        return False
    elif isinstance(eff, Effects.Select):
        return (_is_FV(x, eff.cond) or
                _is_FV(x, eff.tcase) or
                _is_FV(x, eff.fcase))
    elif isinstance(eff, Effects.ConfigField):
        return False
    elif isinstance(eff, Effects.effect):
        return ( any( _is_FV(x, es) for es in eff.reads ) or
                 any( _is_FV(x, es) for es in eff.writes ) or
                 any( _is_FV(x, es) for es in eff.reduces ) or
                 any( _is_FV(x, ce) for ce in eff.config_reads ) or
                 any( _is_FV(x, ce) for ce in eff.config_writes ) )
    else:
        assert False, f"bad case: {type(eff)}"
"""



from enum import Enum, auto
from .LoopIR import LoopIR, T


class ProcedureBase:
    pass


class ExoType(Enum):
    F16 = auto()
    F32 = auto()
    F64 = auto()
    UI8 = auto()
    I8 = auto()
    UI16 = auto()
    I32 = auto()
    R = auto()
    Index = auto()
    Bool = auto()
    Size = auto()
    Int = auto()
    Stride = auto()

    def is_indexable(self):
        return self in [ExoType.Index, ExoType.Size, ExoType.Int, ExoType.Stride]

    def is_numeric(self):
        return self in [
            ExoType.F16,
            ExoType.F32,
            ExoType.F64,
            ExoType.I8,
            ExoType.UI8,
            ExoType.UI16,
            ExoType.I32,
            ExoType.R,
        ]

    def is_bool(self):
        return self == ExoType.Bool


def loopir_type_to_exotype(typ: T) -> ExoType:
    mapping = {
        LoopIR.F16: ExoType.F16,
        LoopIR.F32: ExoType.F32,
        LoopIR.F64: ExoType.F64,
        LoopIR.UINT8: ExoType.UI8,
        LoopIR.INT8: ExoType.I8,
        LoopIR.UINT16: ExoType.UI16,
        LoopIR.INT32: ExoType.I32,
        LoopIR.Num: ExoType.R,
        LoopIR.Index: ExoType.Index,
        LoopIR.Bool: ExoType.Bool,
        LoopIR.Size: ExoType.Size,
        LoopIR.Int: ExoType.Int,
        LoopIR.Stride: ExoType.Stride,
    }
    for key, val in mapping.items():
        if isinstance(typ, key):
            return val
    assert False, f"Type {typ} not found"



from .LoopIR import T, UAST, LoopIR
from .builtins import BuiltIn_Typecheck_Error
from .memory import *


# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #

# This pass converts from UAST to LoopIR
#   In the process, different kinds of expressions are identified and
#   checked for proper usage.
#
#   Non-Goals:
#     - The UAST already has symbol conversion sorted out, so we
#       don't need to do that or to capture additional Python values
#       due to metaprogramming etc.
#     - We're not going to check bounds correctness here
#     - We're not going to check for parallelism-induced race-conditions here
#

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# Helper functions

# --------------------------------------------------------------------------- #
# --------------------------------------------------------------------------- #
# The typechecker


class TypeChecker:
    def __init__(self, proc):
        self.uast_proc = proc
        self.env = dict()
        self.errors = []

        args = []
        for a in proc.args:
            typ = self.check_t(a.type)
            self.env[a.name] = typ
            mem = a.mem
            if mem is None:
                mem = DRAM
            args.append(LoopIR.fnarg(a.name, typ, mem, a.srcinfo))

        preds = []
        for p in proc.preds:
            pred = self.check_e(p, is_index=True)
            if pred.type != T.err and pred.type != T.bool:
                self.err(pred, f"expected a bool expression")
            preds.append(pred)

        body = self.check_stmts(proc.body)

        if not proc.name:
            self.err(proc, "expected all procedures to be named")

        instr = proc.instr
        if instr:
            instr = LoopIR.instr(c_instr=instr.c_instr, c_global=instr.c_global)

        self.loopir_proc = LoopIR.proc(
            name=proc.name or "anon",
            args=args,
            preds=preds,
            body=body,
            instr=instr,
            eff=None,
            srcinfo=proc.srcinfo,
        )

        # do error checking here
        if len(self.errors) > 0:
            raise TypeError(
                "Errors occurred during typechecking:\n" + "\n".join(self.errors)
            )

    def get_loopir(self):
        return self.loopir_proc

    def err(self, node, msg):
        self.errors.append(f"{node.srcinfo}: {msg}")

    def check_stmts(self, body):
        assert len(body) > 0 or self.uast_proc.instr
        stmts = []
        for s in body:
            stmts += self.check_single_stmt(s)
        return stmts

    def check_access(self, node, nm, idx, lvalue=False):
        # check indexing
        idx = [self.check_e(i, is_index=True) for i in idx]
        for i in idx:
            if i.type != T.err and not i.type.is_indexable():
                self.err(i, f"cannot index with expression of type '{i.type}'")

        # check compatibility with buffer type
        typ = self.env[nm]
        if typ is T.err:
            pass
        elif typ.is_numeric():
            if len(idx) > len(typ.shape()):
                self.err(
                    node,
                    f"expected access of variable "
                    f"'{nm}' of type '{typ}' to have "
                    f"no more than {len(typ.shape())} "
                    f"indices, but {len(idx)} were found.",
                )
                typ = T.err
            elif lvalue and len(typ.shape()) != len(idx):
                self.err(
                    node,
                    f"expected lvalue access of variable "
                    f"'{nm}' of type '{typ}' to have "
                    f"exactly {len(typ.shape())} "
                    f"indices, but {len(idx)} were found.",
                )
                typ = T.err
            elif len(idx) > 0:
                assert len(idx) == len(typ.shape())
                typ = typ.basetype()
            # else if len(idx) == 0 then just fall through

        elif lvalue:
            self.err(
                node,
                f"cannot assign/reduce to '{nm}', "
                f"a non-numeric variable of type '{typ}'",
            )
            typ = T.err
        elif len(idx) > 0:
            self.err(node, f"cannot index a variable of type '{typ}'")

        return idx, typ

    def check_single_stmt(self, stmt):
        if isinstance(stmt, UAST.FreshAssign):
            rhs = self.check_e(stmt.rhs)

            # We see a statement of the form
            #   nm = ...
            # we have a windowing statement
            #   nm = x[...]
            # It doesn't make sense to have a general freshassign,
            # when we don't know which memory to allocate.
            if rhs.type == T.err:
                self.env[stmt.name] = T.err
                return []
            elif isinstance(rhs.type, T.Window):
                assert isinstance(rhs, LoopIR.WindowExpr)
                self.env[stmt.name] = rhs.type
                return [LoopIR.WindowStmt(stmt.name, rhs, None, stmt.srcinfo)]
            else:
                self.err(
                    stmt,
                    f"unable to disambiguate assignment to "
                    f"undefined variable '{stmt.name}'",
                )
                self.env[stmt.name] = T.err
                return []

        if isinstance(stmt, (UAST.Assign, UAST.Reduce)):
            rhs = self.check_e(stmt.rhs)
            if rhs.type != T.err and not rhs.type.is_real_scalar():
                self.err(rhs, f"cannot assign/reduce a '{rhs.type}' type value")

            idx, typ = self.check_access(stmt, stmt.name, stmt.idx, lvalue=True)
            assert typ.is_real_scalar() or typ is T.err

            IRnode = LoopIR.Assign if isinstance(stmt, UAST.Assign) else LoopIR.Reduce
            return [IRnode(stmt.name, typ, idx, rhs, None, stmt.srcinfo)]

        elif isinstance(stmt, UAST.WriteConfig):
            # Check that field is in config
            if not stmt.config.has_field(stmt.field):
                self.err(
                    stmt.field,
                    f"expected '{stmt.field}'  to be a field "
                    f"in config '{stmt.config.name()}'",
                )

            ftyp = stmt.config.lookup(stmt.field)[1]
            rhs = self.check_e(
                stmt.rhs,
                is_index=ftyp.is_indexable() or ftyp.is_stridable() or ftyp == T.bool,
            )

            if rhs.type != T.err:
                if ftyp.is_real_scalar():
                    if not rhs.type.is_real_scalar():
                        self.err(
                            rhs,
                            f"expected a real scalar value, but "
                            f"got an expression of type {rhs.type}",
                        )
                elif ftyp.is_indexable():
                    if not rhs.type.is_indexable():
                        self.err(
                            rhs,
                            f"expected an index or size type "
                            f"expression, but got type {rhs.type}",
                        )
                elif ftyp == T.bool:
                    if rhs.type != T.bool:
                        self.err(
                            rhs,
                            f"expected a bool expression, but got type {rhs.type}",
                        )
                elif ftyp.is_stridable():
                    if not rhs.type.is_stridable():
                        self.err(
                            rhs,
                            f"expected a stride type expression, "
                            f"but got type {rhs.type}",
                        )
                else:
                    assert False, "bad case"

            return [
                LoopIR.WriteConfig(stmt.config, stmt.field, rhs, None, stmt.srcinfo)
            ]
        elif isinstance(stmt, UAST.Pass):
            return [LoopIR.Pass(None, stmt.srcinfo)]

        elif isinstance(stmt, UAST.If):
            cond = self.check_e(stmt.cond, is_index=True)
            if cond.type != T.err and cond.type != T.bool:
                self.err(cond, f"expected a bool expression")
            body = self.check_stmts(stmt.body)
            ebody = []
            if len(stmt.orelse) > 0:
                ebody = self.check_stmts(stmt.orelse)
            return [LoopIR.If(cond, body, ebody, None, stmt.srcinfo)]

        elif isinstance(stmt, UAST.For):
            self.env[stmt.iter] = T.index

            # handle standard ParRanges
            parerr = (
                "currently supporting for-loops of the form:\n"
                "  'for _ in par(affine_expression, affine_expression):' and "
                "'for _ in seq(affine_expression, affine_expression):'"
            )

            if not isinstance(stmt.cond, (UAST.ParRange, UAST.SeqRange)):
                self.err(stmt.cond, parerr)

            lo = self.check_e(stmt.cond.lo, is_index=True)
            if lo.type != T.err and not lo.type.is_indexable():
                self.err(lo, "expected loop bound to be indexable.")
            hi = self.check_e(stmt.cond.hi, is_index=True)
            if hi.type != T.err and not hi.type.is_indexable():
                self.err(hi, "expected loop bound to be indexable.")

            body = self.check_stmts(stmt.body)
            if isinstance(stmt.cond, UAST.SeqRange):
                return [
                    LoopIR.For(
                        stmt.iter, lo, hi, body, LoopIR.Seq(), None, stmt.srcinfo
                    )
                ]
            elif isinstance(stmt.cond, UAST.ParRange):
                return [
                    LoopIR.For(
                        stmt.iter, lo, hi, body, LoopIR.Par(), None, stmt.srcinfo
                    )
                ]
            else:
                assert False, "bad case"

        elif isinstance(stmt, UAST.Alloc):
            typ = self.check_t(stmt.type)
            self.env[stmt.name] = typ
            mem = stmt.mem
            if mem is None:
                mem = DRAM
            return [LoopIR.Alloc(stmt.name, typ, mem, None, stmt.srcinfo)]

        elif isinstance(stmt, UAST.Call):
            args = [
                self.check_e(
                    call_a, is_index=sig_a.type in (T.size, T.index, T.stride, T.bool)
                )
                for call_a, sig_a in zip(stmt.args, stmt.f.args)
            ]

            for call_a, sig_a in zip(args, stmt.f.args):
                if call_a.type == T.err:
                    pass
                elif sig_a.type is T.size or sig_a.type is T.index:
                    if not call_a.type.is_indexable():
                        self.err(
                            call_a,
                            "expected size or index type "
                            "expression, "
                            f"but got type {call_a.type}",
                        )

                elif sig_a.type is T.bool:
                    if not call_a.type is T.bool:
                        self.err(
                            call_a,
                            "expected bool-type variable, "
                            f"but got type {call_a.type}",
                        )

                elif sig_a.type is T.stride:
                    if not call_a.type.is_stridable():
                        self.err(
                            call_a,
                            "expected stride-type variable, "
                            f"but got type {call_a.type}",
                        )

                elif sig_a.type.is_numeric():
                    if len(call_a.type.shape()) != len(sig_a.type.shape()):
                        self.err(
                            call_a,
                            f"expected argument of type '{sig_a.type}', "
                            f"but got '{call_a.type}'",
                        )

                    # ensure scalars are simply variable names
                    elif call_a.type.is_real_scalar():
                        if not isinstance(call_a, LoopIR.ReadConfig) and not (
                            isinstance(call_a, LoopIR.Read) and len(call_a.idx) == 0
                        ):
                            self.err(
                                call_a,
                                "expected scalar arguments "
                                "to be simply variable names "
                                "for now",
                            )

                else:
                    assert False, "bad argument type case"

            return [LoopIR.Call(stmt.f, args, None, stmt.srcinfo)]
        else:
            assert False, "not a loopir in check_stmts"

    def check_w_access(self, e, orig_hi):
        if isinstance(e, UAST.Point):
            pt = self.check_e(e.pt, is_index=True)
            if pt.type != T.err and not pt.type.is_indexable():
                self.err(pt, f"cannot index with expression of type '{pt.type}'")
            return LoopIR.Point(pt, e.srcinfo)

        elif isinstance(e, UAST.Interval):
            if e.lo is None:
                lo = LoopIR.Const(0, T.int, e.srcinfo)
            else:
                lo = self.check_e(e.lo, is_index=True)
                if lo.type != T.err and not lo.type.is_indexable():
                    self.err(lo, f"cannot index with expression of type '{lo.type}'")

            if e.hi is None:
                hi = orig_hi
            else:
                hi = self.check_e(e.hi, is_index=True)
                if hi.type != T.err and not hi.type.is_indexable():
                    self.err(hi, f"cannot index with expression of type '{hi.type}'")

            return LoopIR.Interval(lo, hi, e.srcinfo)

    def build_window_shape(self, ws):
        def subtract(hi, lo):
            if isinstance(lo, LoopIR.Const) and lo.val == 0:
                return hi
            else:
                return LoopIR.BinOp("-", hi, lo, T.index, hi.srcinfo)

        return [subtract(w.hi, w.lo) for w in ws if isinstance(w, LoopIR.Interval)]

    def check_e(self, e, is_index=False):
        if isinstance(e, UAST.Read):
            typ = self.env[e.name]
            # if we only partially accessed the base tensor/window,
            # then this is sugar for a windowing expression
            if typ.is_tensor_or_window() and 0 < len(e.idx) < len(typ.shape()):
                # expand to a windowing expression
                # x[i] == x[i,:]

                idxs = [UAST.Point(i, i.srcinfo) for i in e.idx] + [
                    UAST.Interval(None, None, e.srcinfo)
                    for _ in range(0, len(typ.shape()) - len(e.idx))
                ]

                desugared = UAST.WindowExpr(e.name, idxs, e.srcinfo)
                return self.check_e(desugared)

            # otherwise, we have a normal access
            else:
                idx, typ = self.check_access(e, e.name, e.idx, lvalue=False)
                return LoopIR.Read(e.name, idx, typ, e.srcinfo)

        elif isinstance(e, UAST.WindowExpr):
            typ = self.env[e.name]
            if not typ.is_tensor_or_window():
                self.err(
                    e,
                    f"cannot perform windowing on non-tensor, "
                    f"non-window type {e.base}",
                )
                return LoopIR.WindowExpr(e.name, [], T.err, e.srcinfo)

            shape = typ.shape()
            if len(shape) != len(e.idx):
                self.err(
                    e,
                    f"expected {len(shape)} indices for window "
                    f"but got {len(e.idx)}",
                )

            idx = [self.check_w_access(w, t) for w, t in zip(e.idx, shape)]

            # TODO: Construct as_tensor...
            window_shape = self.build_window_shape(idx)
            as_tensor = T.Tensor(window_shape, True, typ.type)

            w_typ = T.Window(typ, as_tensor, e.name, idx)
            return LoopIR.WindowExpr(e.name, idx, w_typ, e.srcinfo)

        elif isinstance(e, UAST.Const):
            ty = {float: T.R, bool: T.bool, int: T.int if is_index else T.R}.get(
                type(e.val)
            )
            if not ty:
                self.err(
                    e,
                    f"literal of unexpected type '{type(e.val)}' and "
                    f"value: {e.val}",
                )
                return LoopIR.Const(0, T.err, e.srcinfo)
            return LoopIR.Const(e.val, ty, e.srcinfo)

        elif isinstance(e, UAST.USub):
            arg = self.check_e(e.arg, is_index=is_index)
            if arg.type.is_real_scalar() or arg.type.is_indexable():
                if isinstance(arg, LoopIR.Const):
                    return LoopIR.Const(-arg.val, arg.type, e.srcinfo)
                else:
                    return LoopIR.USub(arg, arg.type, e.srcinfo)
            elif arg.type != T.err:
                self.err(e, f"cannot negate expression of type '{arg.type}'")
            return LoopIR.Const(0, T.err, e.srcinfo)

        elif isinstance(e, UAST.BinOp):
            lhs = self.check_e(e.lhs, is_index=is_index)
            rhs = self.check_e(e.rhs, is_index=is_index)
            typ = T.err
            if lhs.type == T.err or rhs.type == T.err:
                typ = T.err
            elif e.op in ("and", "or"):
                for operand in (lhs, rhs):
                    if operand.type is not T.bool:
                        self.err(operand, "expected 'bool' argument to logical op")
                typ = T.bool
            elif e.op == "==" and (
                (lhs.type == T.bool and rhs.type == T.bool)
                or (
                    lhs.type.is_stridable()
                    and rhs.type.is_stridable()
                    and not (lhs.type == T.int and rhs.type == T.int)
                )
            ):
                typ = T.bool
            elif e.op in ("<", "<=", "==", ">", ">="):
                for operand in (lhs, rhs):
                    if not operand.type.is_indexable():
                        self.err(
                            operand,
                            f"expected 'index' or 'size' argument to "
                            f"comparison op: {e.op}",
                        )
                typ = T.bool
            elif e.op in ("+", "-", "*", "/", "%"):
                if lhs.type.is_real_scalar():
                    if not rhs.type.is_real_scalar():
                        self.err(rhs, "expected scalar type")
                        typ = T.err
                    elif e.op == "%":
                        self.err(e, "cannot compute modulus of 'R' values")
                        typ = T.err
                    else:
                        typ = lhs.type
                elif rhs.type.is_real_scalar():
                    self.err(lhs, "expected scalar type")
                elif lhs.type == T.bool or rhs.type == T.bool:
                    node = lhs if lhs.type == T.bool else rhs
                    self.err(node, "cannot perform arithmetic on 'bool' values")
                    typ = T.err
                elif lhs.type == T.stride or rhs.type == T.stride:
                    node = lhs if lhs.type == T.bool else rhs
                    self.err(node, "cannot perform arithmetic on 'stride' values")
                    typ = T.err
                elif lhs.type.is_tensor_or_window() or rhs.type.is_tensor_or_window():
                    self.err(lhs, "cannot perform arithmetic on tensors")
                    typ = T.err
                else:
                    assert lhs.type.is_indexable()
                    assert rhs.type.is_indexable()
                    if e.op == "/" or e.op == "%":
                        if rhs.type != T.int or not isinstance(rhs, LoopIR.Const):
                            self.err(
                                rhs,
                                "cannot divide or modulo by a " "non-constant value",
                            )
                            typ = T.err
                        elif rhs.val <= 0:
                            self.err(
                                rhs,
                                "cannot divide or modulo by zero "
                                "or a negative value",
                            )
                            typ = T.err

                        typ = lhs.type
                    elif e.op == "*":
                        if lhs.type == T.int:
                            typ = rhs.type
                        elif rhs.type == T.int:
                            typ = lhs.type
                        else:
                            self.err(
                                e,
                                "cannot multiply two non-constant "
                                "indexing/sizing expressions, since "
                                "the result would be non-affine",
                            )
                            typ = T.err
                    else:  # + or -
                        if lhs.type == T.index or rhs.type == T.index:
                            typ = T.index
                        elif lhs.type == T.size or rhs.type == T.size:
                            typ = T.size
                        else:
                            typ = T.int

            else:
                assert False, f"bad op: '{e.op}'"

            return LoopIR.BinOp(e.op, lhs, rhs, typ, e.srcinfo)

        elif isinstance(e, UAST.BuiltIn):

            args = [self.check_e(a) for a in e.args]

            try:
                typ = e.f.typecheck(args)
            except BuiltIn_Typecheck_Error as err:
                typ = T.err
                self.err(e, str(err))

            return LoopIR.BuiltIn(e.f, args, typ, e.srcinfo)

        elif isinstance(e, UAST.StrideExpr):
            idx, typ = self.check_access(e, e.name, [], lvalue=False)
            assert len(idx) == 0
            if typ == T.err:
                pass
            elif not typ.is_tensor_or_window():
                self.err(e, f"expected {e.name} to be a tensor or window")
            else:
                shape = typ.shape()
                if not (0 <= e.dim < len(shape)):
                    self.err(
                        e,
                        f"expected {e.dim} to be in-bounds "
                        f"(i.e. 0 <= {e.dim} < {len(shape)})",
                    )

            return LoopIR.StrideExpr(e.name, e.dim, T.stride, e.srcinfo)

        elif isinstance(e, UAST.ParRange):
            assert False, (
                "parser should not place ParRange anywhere "
                "outside of a for-loop condition"
            )
        elif isinstance(e, UAST.ReadConfig):
            if not e.config.has_field(e.field):
                self.err(
                    e.field,
                    f"'{e.field}' has to be a field in config '{e.config.name()}'",
                )

            ftyp = e.config.lookup(e.field)[1]
            return LoopIR.ReadConfig(e.config, e.field, ftyp, e.srcinfo)
        else:
            assert False, "not a LoopIR in check_e"

    _typ_table = {
        UAST.Num: T.R,
        UAST.F16: T.f16,
        UAST.F32: T.f32,
        UAST.F64: T.f64,
        UAST.INT8: T.int8,
        UAST.UINT8: T.uint8,
        UAST.UINT16: T.uint16,
        UAST.INT32: T.int32,
        UAST.Bool: T.bool,
        UAST.Int: T.int,
        UAST.Size: T.size,
        UAST.Index: T.index,
        UAST.Stride: T.stride,
    }

    def check_t(self, typ):
        if type(typ) in TypeChecker._typ_table:
            return TypeChecker._typ_table[type(typ)]
        elif isinstance(typ, UAST.Tensor):
            hi = [self.check_e(h, is_index=True) for h in typ.hi]
            sub_typ = self.check_t(typ.type)
            for h in hi:
                if not h.type.is_indexable():
                    self.err(
                        h,
                        "expected array size expression "
                        "to have type 'size' or type 'index'",
                    )
            return T.Tensor(hi, typ.is_window, sub_typ)
        else:
            assert False, "bad case"
